import aiohttp
from aiohttp.test_utils import unittest_run_loop
from tests.cbpi_config_fixture import CraftBeerPiTestCase

# class WebSocketTestCase(CraftBeerPiTestCase):

#     @unittest_run_loop
#     async def test_brewing_process(self):

#         count_step_done = 0
#         async with self.client.ws_connect('/ws') as ws:
#             await ws.send_json(data=dict(topic="step/stop"))
#             await ws.send_json(data=dict(topic="step/start"))
#             async for msg in ws:
#                 if msg.type == aiohttp.WSMsgType.TEXT:
#                     try:
#                         msg_obj = msg.json()
#                         topic = msg_obj.get("topic")
#                         if topic == "job/step/done":
#                             count_step_done = count_step_done + 1
#                         if topic == "step/brewing/finished":
#                             await ws.send_json(data=dict(topic="close"))
#                     except Exception as e:
#                         print(e)
#                         break
#                 elif msg.type == aiohttp.WSMsgType.ERROR:
#                     break

#         assert count_step_done == 4

#     @unittest_run_loop
#     async def test_wrong_format(self):

#         async with self.client.ws_connect('/ws') as ws:
#             await ws.send_json(data=dict(a="close"))
#             async for msg in ws:
#                 print("MSG TYP", msg.type, msg.data)
#                 if msg.type == aiohttp.WSMsgType.TEXT:
#                     msg_obj = msg.json()
#                     if msg_obj["topic"] != "connection/success":
#                         print(msg.data)
#                         raise Exception()

#                 else:
#                     raise Exception()




__all__ = ["PropertyType", "Property"]

class PropertyType(object):
    pass

class Property(object):
    class Select(PropertyType):

        '''
        Select Property. The user can select value from list set as options parameter
        '''

        def __init__(self, label, options, description=""):
            '''
            
            :param label: 
            :param options: 
            :param description: 
            '''
            PropertyType.__init__(self)
            self.label = label
            self.options = options
            self.description = description

    class Number(PropertyType):

        '''
        The user can set a number value 
        '''
        def __init__(self, label, configurable=False, default_value=None, unit="", description=""):
            '''
            
            :param label: 
            :param configurable: 
            :param default_value: 
            :param unit: 
            :param description: 
            '''
            PropertyType.__init__(self)
            self.label = label
            self.configurable = configurable
            self.default_value = default_value
            self.description = description

    class Text(PropertyType):

        '''
        The user can set a text value 
        '''
        def __init__(self, label, configurable=False, default_value="", description=""):
            '''
            
            :param label: 
            :param configurable: 
            :param default_value: 
            :param description: 
            '''
            PropertyType.__init__(self)
            self.label = label
            self.configurable = configurable
            self.default_value = default_value
            self.description = description

    class Actor(PropertyType):

        '''
        The user select an actor which is available in the system. The value of this variable will be the actor id
        '''
        def __init__(self, label, description=""):
            '''
            
            :param label: 
            :param description: 
            '''
            PropertyType.__init__(self)
            self.label = label
            self.configurable = True
            self.description = description

    class Sensor(PropertyType):
        '''
        The user select a sensor which is available in the system. The value of this variable will be the sensor id
        '''

        def __init__(self, label, description=""):
            '''
            
            :param label: 
            :param description: 
            '''
            PropertyType.__init__(self)
            self.label = label
            self.configurable = True
            self.description = description

    class Kettle(PropertyType):
        '''
        The user select a kettle which is available in the system. The value of this variable will be the kettle id
        '''

        def __init__(self, label, description=""):
            '''
            
            :param label: 
            :param description: 
            '''

            PropertyType.__init__(self)
            self.label = label
            self.configurable = True
            self.description = description

    class Fermenter(PropertyType):
        '''
        The user select a fermenter which is available in the system. The value of this variable will be the fermenter id
        '''

        def __init__(self, label, description=""):
            '''
            
            :param label: 
            :param description: 
            '''

            PropertyType.__init__(self)
            self.label = label
            self.configurable = True
            self.description = description




import asyncio
import json
from cbpi.api import parameters, Property, CBPiActor
from cbpi.api import *

@parameters([Property.Text(label="Topic", configurable=True, description = "MQTT Topic")])
class MQTTActor(CBPiActor):

    # Custom property which can be configured by the user
    @action("Set Power", parameters=[Property.Number(label="Power", configurable=True, description="Power Setting [0-100]")])
    async def setpower(self,Power = 100 ,**kwargs):
        self.power=int(Power)
        if self.power < 0:
            self.power = 0
        if self.power > 100:
            self.power = 100
        await self.set_power(self.power)

    def __init__(self, cbpi, id, props):
        super(MQTTActor, self).__init__(cbpi, id, props)

    async def on_start(self):
         self.topic = self.props.get("Topic", None)
         self.power = 100
         await self.off()
         self.state = False

    async def on(self, power=None):
        if power is not None:
            if power != self.power:
                power = min(100, power)
                power = max(0, power)
                self.power = round(power)
        await self.cbpi.satellite.publish(self.topic, json.dumps(
            {"state": "on", "power": self.power}), True)
        self.state = True
        pass

    async def off(self):
        self.state = False
        await self.cbpi.satellite.publish(self.topic, json.dumps(
            {"state": "off", "power": self.power}), True)
        pass

    async def run(self):
        while self.running:
            await asyncio.sleep(1)

    def get_state(self):
        return self.state

    async def set_power(self, power):
        self.power = round(power)
        if self.state == True:
            await self.on(power)
        else:
            await self.off()
        await self.cbpi.actor.actor_update(self.id,power)
        pass


import os
import pathlib
import aiohttp
from aiohttp import web
import asyncio
from cbpi.api import *
import xml.etree.ElementTree
import sqlite3
from voluptuous.schema_builder import message
from cbpi.api.dataclasses import NotificationAction, NotificationType, Actor, Sensor, Kettle
from cbpi.controller.kettle_controller import KettleController
from cbpi.api.base import CBPiBase
from cbpi.api.config import ConfigType
import webbrowser
import logging
import os.path
from os import listdir
from os.path import isfile, join
import json
import shortuuid
import yaml
from ..api.step import StepMove, StepResult, StepState
import re
import base64


class UploadController:

    def __init__(self, cbpi):
        self.cbpi = cbpi
        self.logger = logging.getLogger(__name__)
   
    async def get_kbh_recipes(self):
        try:
            path = self.cbpi.config_folder.get_upload_file("kbh.db")
            conn = sqlite3.connect(path)
            c = conn.cursor()
            c.execute('SELECT ID, Sudname, Status FROM Sud')
            data = c.fetchall()
            result = []
            for row in data:
                element = {'value': str(row[0]), 'label': str(row[1])}
                result.append(element)
            return result
        except:
            return []

    async def get_xml_recipes(self):
        try:
            path = self.cbpi.config_folder.get_upload_file("beer.xml")
            e = xml.etree.ElementTree.parse(path).getroot()
            result =[] 
            counter = 1
            for idx, val in enumerate(e.findall('RECIPE')):
                element = {'value': str(counter), 'label': val.find("NAME").text}
                result.append(element)
                counter +=1
            return result
        except:
            return []

    async def get_json_recipes(self):
        try:
            path = self.cbpi.config_folder.get_upload_file("mmum.json")
            e = json.load(open(path))
            result =[] 
            result.append({'value': str(1), 'label': e['Name']})
            return result
        except:
            return []

    async def get_brewfather_recipes(self,offset=0):
        brewfather = True
        result=[]
        self.url="https://api.brewfather.app/v1/recipes"
        brewfather_user_id = self.cbpi.config.get("brewfather_user_id", None)
        if brewfather_user_id == "" or brewfather_user_id is None:
            brewfather = False

        brewfather_api_key = self.cbpi.config.get("brewfather_api_key", None)
        if brewfather_api_key == "" or brewfather_api_key is None:
            brewfather = False

        if brewfather == True:
            encodedData = base64.b64encode(bytes(f"{brewfather_user_id}:{brewfather_api_key}", "ISO-8859-1")).decode("ascii")
            headers={"Authorization": "Basic %s" % encodedData}
            parameters={"limit": 50, 'offset': offset}
            async with aiohttp.ClientSession(headers=headers) as bf_session:
                async with bf_session.get(self.url, params=parameters) as r:
                    bf_recipe_list = await r.json()
                await bf_session.close()

            if bf_recipe_list:
                for row in bf_recipe_list:
                    recipe_id = row['_id']
                    name = row['name']
                    element = {'value': recipe_id, 'label': name}
                    result.append(element)
                return result
            else:
                return []

        else:
            return []

        
    def get_creation_path(self):
        creation_path = self.cbpi.config.get("RECIPE_CREATION_PATH", "upload")
        path = {'path': 'upload'} if creation_path == '' else {'path': creation_path}
        return path

    def allowed_file(self, filename, extension):
        return '.' in filename and filename.rsplit('.', 1)[1] in set([extension])


    async def FileUpload(self, data):
        fileData = data['File']
        filename = fileData.filename
        recipe_file = fileData.file
        content_type = fileData.content_type

        if content_type == 'text/xml':
            try:
                beer_xml = recipe_file.read().decode('utf-8','replace')
                if recipe_file and self.allowed_file(filename, 'xml'):
                    self.path = self.cbpi.config_folder.get_upload_file("beer.xml")
    
                    f = open(self.path, "w")
                    f.write(beer_xml)
                    f.close()
                    self.cbpi.notify("Success", "XML Recipe {} has been uploaded".format(filename), NotificationType.SUCCESS)
            except Exception as e:
                self.cbpi.notify("Error" "XML Recipe upload failed: {}".format(e), NotificationType.ERROR)
                pass

        elif content_type == 'application/json':
            try:
                mmum_json = recipe_file.read().decode('utf-8','replace')
                if recipe_file and self.allowed_file(filename, 'json'):
                    self.path = self.cbpi.config_folder.get_upload_file("mmum.json")
    
                    f = open(self.path, "w")
                    f.write(mmum_json)
                    f.close()
                    self.cbpi.notify("Success", "JSON Recipe {} has been uploaded".format(filename), NotificationType.SUCCESS)
            except Exception as e:
                self.cbpi.notify("Error" "JSON Recipe upload failed: {}".format(e), NotificationType.ERROR)
                pass

        elif content_type == 'application/octet-stream':
            try:
                content = recipe_file.read()
                if recipe_file and self.allowed_file(filename, 'sqlite'):
                    self.path = self.cbpi.config_folder.get_upload_file("kbh.db")

                    f=open(self.path, "wb")
                    f.write(content)
                    f.close()
                    self.cbpi.notify("Success", "Kleiner Brauhelfer database has been uploaded", NotificationType.SUCCESS)

            except:
                self.cbpi.notify("Error", "Kleiner Brauhelfer database upload failed", NotificationType.ERROR)
                pass
        else:
            self.cbpi.notify("Error", "Wrong content type. Upload failed", NotificationType.ERROR)

    async def kbh_recipe_creation(self, Recipe_ID):
        config = self.get_config_values()
        if self.kettle is not None:
            # load beerxml file located in upload folder
            self.path = self.cbpi.config_folder.get_upload_file("kbh.db")
            if os.path.exists(self.path) is False:
                self.cbpi.notify("File Not Found", "Please upload a kbh V2 databsel file", NotificationType.ERROR)
                
            try:
                # Get Recipe Nmae
                conn = sqlite3.connect(self.path)
                c = conn.cursor()
                c.execute('SELECT Sudname FROM Sud WHERE ID = ?', (Recipe_ID,))
                row = c.fetchone()
                name = row[0]

                # get MashIn Temp
                mashin_temp = None
                c.execute('SELECT Temp FROM Rasten WHERE Typ = 0 AND SudID = ?', (Recipe_ID,))
                row = c.fetchone()
                try:
                    if self.cbpi.config.get("TEMP_UNIT", "C") == "C":
                        mashin_temp = str(int(row[0]))
                    else:
                        mashin_temp = str(round(9.0 / 5.0 * int(row[0]) + 32))
                except:
                    pass

                # get the hop addition times
                c.execute('SELECT Zeit FROM Hopfengaben WHERE Vorderwuerze = 0 AND SudID = ?', (Recipe_ID,))
                hops = c.fetchall()

                # get the misc addition times
                c.execute('SELECT Zugabedauer FROM WeitereZutatenGaben WHERE Zeitpunkt = 1 AND SudID = ?', (Recipe_ID,))
                miscs = c.fetchall()

                try:
                    c.execute('SELECT Zeit FROM Hopfengaben WHERE Vorderwuerze = 1 AND SudID = ?', (Recipe_ID,))
                    FW_Hops = c.fetchall()
                    FirstWort = self.getFirstWort(FW_Hops,"kbh")
                except:
                    FirstWort = "No"

                c.execute('SELECT Kochdauer FROM Sud WHERE ID = ?', (Recipe_ID,))
                row = c.fetchone()
                BoilTime = str(int(row[0]))



                await self.create_recipe(name)

                if mashin_temp is not None:
                    step_type = self.mashin if self.mashin != "" else "MashInStep"
                    step_string = { "name": "MashIn",
                                    "props": {
                                        "AutoMode": self.AutoMode,
                                        "Kettle": self.id,
                                        "Sensor": self.kettle.sensor,
                                        "Temp": mashin_temp,
                                        "Timer": "0",
                                        "Notification": "Target temperature reached. Please add malt."
                                        },
                                    "status_text": "",
                                    "status": "I",
                                    "type": step_type
                                    }
                    await self.create_step(step_string)

                for row in c.execute('SELECT Name, Temp, Dauer FROM Rasten WHERE Typ <> 0 AND SudID = ?', (Recipe_ID,)):
                    if mashin_temp is None and self.addmashin == "Yes":
                        step_type = self.mashin if self.mashin != "" else "MashInStep"
                        step_string = { "name": "MashIn",
                                    "props": {
                                        "AutoMode": self.AutoMode,
                                        "Kettle": self.id,
                                        "Sensor": self.kettle.sensor,
                                        "Temp": str(int(row[1])) if self.TEMP_UNIT == "C" else str(round(9.0 / 5.0 * int(row[1]) + 32)),
                                        "Timer": "0",
                                        "Notification": "Target temperature reached. Please add malt."
                                        },
                                    "status_text": "",
                                    "status": "I",
                                    "type": step_type
                                    }
                        await self.create_step(step_string)


                    step_type = self.mash if self.mash != "" else "MashStep"
                    step_string = { "name": str(row[0]),
                                    "props": {
                                        "AutoMode": self.AutoMode,
                                        "Kettle": self.id,
                                        "Sensor": self.kettle.sensor,
                                        "Temp": str(int(row[1])) if self.TEMP_UNIT == "C" else str(round(9.0 / 5.0 * int(row[1]) + 32)),
                                        "Timer": str(int(row[2]))
                                        },
                                    "status_text": "",
                                    "status": "I",
                                    "type": step_type
                                    }
                    await self.create_step(step_string)

                # MashOut -> Notification step that sends notification and waits for user input to move to next step (AutoNext=No)
                if self.mashout == "NotificationStep":
                    step_string = { "name": "Lautering",
                                    "props": {
                                        "AutoNext": "No",
                                        "Kettle": self.id,
                                        "Notification": "Mash Process completed. Please start lautering and press next to start boil."

                                        },
                                    "status_text": "",
                                    "status": "I",
                                    "type": self.mashout
                                    }
                    await self.create_step(step_string)


                Hops = self.getBoilAlerts(hops, miscs, "kbh")
                step_type = self.boil if self.boil != "" else "BoilStep"
                step_string = { "name": "Boil Step",
                            "props": {
                                "AutoMode": self.AutoMode,
                                "Kettle": self.boilid,
                                "Sensor": self.boilkettle.sensor,
                                "Temp": int(self.BoilTemp),
                                "Timer": BoilTime,
                                "First_Wort": FirstWort,
                                "LidAlert": "Yes",
                                "Hop_1": Hops[0],
                                "Hop_2": Hops[1],
                                "Hop_3": Hops[2],
                                "Hop_4": Hops[3],
                                "Hop_5": Hops[4],
                                "Hop_6": Hops[5]
                                },
                            "status_text": "",
                            "status": "I",
                            "type": step_type 
                        }

                await self.create_step(step_string)

                await self.create_Whirlpool_Cooldown()
 
                self.cbpi.notify('KBH Recipe created', name, NotificationType.INFO)

            except:
                self.cbpi.notify('KBH Recipe creation failure', name, NotificationType.ERROR)
                pass
        else:
            self.cbpi.notify('Recipe Upload', 'No default Kettle defined. Please specify default Kettle in settings', NotificationType.ERROR)

    def findMax(self, string):
        self.path = self.cbpi.config_folder.get_upload_file("mmum.json")
        e = json.load(open(self.path))
        for idx in range(1,20):
            search_string = string.replace("%%",str(idx))
            i = idx
            if search_string not in e:
                break
        return i

    def getJsonMashin(self, id):
        self.path = self.cbpi.config_folder.get_upload_file("mmum.json")
        e = json.load(open(self.path))
        return float(e['Infusion_Einmaischtemperatur'])

    async def json_recipe_creation(self, Recipe_ID):
        config = self.get_config_values()

        if self.kettle is not None:
            # load mmum-json file located in upload folder
            self.path = self.cbpi.config_folder.get_upload_file("mmum.json")
            if os.path.exists(self.path) is False:
                self.cbpi.notify("File Not Found", "Please upload a MMuM-JSON File", NotificationType.ERROR)

            e = json.load(open(self.path))
            name = e['Name']
            boil_time = float(e['Kochzeit_Wuerze'])
            
            await self.create_recipe(name)
            
            
            hops = []
            for idx in range(1,self.findMax("Hopfen_%%_Kochzeit")):
                hops_name = "%sg %s %s%% alpha" % (e["Hopfen_{}_Menge".format(idx)],e["Hopfen_{}_Sorte".format(idx)],e["Hopfen_{}_alpha".format(idx)])
                if e["Hopfen_{}_Kochzeit".format(idx)].isnumeric():
                    if boil_time is not e["Hopfen_{}_Kochzeit".format(idx)].isnumeric():
                        alert = float(e["Hopfen_{}_Kochzeit".format(idx)])
                elif e["Hopfen_{}_Kochzeit".format(idx)] == "Whirlpool":
                    alert = float(1)
                else:
                    self.api.notify(headline="No Number at Hoptime", message="Please change json-File at Hopfen_{}_Kochzeit".format(idx), type="danger")
                    alert = float(1)
                hops.append({"name":hops_name,"time":alert})
                
            
            firstHops=[]
            for idx in range(1,self.findMax("Hopfen_VWH_%%_Sorte")):
                firstHops_name = "%sg %s %s%% alpha" % (e["Hopfen_VWH_{}_Menge".format(idx)],e["Hopfen_VWH_{}_Sorte".format(idx)],e["Hopfen_VWH_{}_alpha".format(idx)])
            
                firstHops.append({"name":firstHops_name})
            
            
            FirstWort= self.getFirstWort(firstHops, "json")
            
            miscs = []
            for idx in range(1,self.findMax("WeitereZutat_Wuerze_%%_Kochzeit")):
                miscs_name = "%s%s %s" % (e["WeitereZutat_Wuerze_{}_Menge".format(idx)],e["WeitereZutat_Wuerze_{}_Einheit".format(idx)],e["WeitereZutat_Wuerze_{}_Name".format(idx)])
                if e["WeitereZutat_Wuerze_{}_Kochzeit".format(idx)].isnumeric():
                    alert = float(e["WeitereZutat_Wuerze_{}_Kochzeit".format(idx)])
                elif e["WeitereZutat_Wuerze_{}_Kochzeit".format(idx)] == "Whirlpool":
                    alert = float(1)
                else:
                    self.api.notify(headline="No Number at Hoptime", message="Please change json-File at WeitereZutat_Wuerze_{}_Kochzeit".format(idx), type="danger")
                    alert = float(1)
                miscs.append({"name":miscs_name,"time":alert})
                
                
            # Mash Steps -> first step is different as it heats up to defined temp and stops with notification to add malt
            # AutoMode is yes to start and stop automatic mode or each step
            MashIn_Flag = True
            step_kettle = self.id
            last_step_temp = 0
            logging.info(step_kettle)  ###################################################
            for row in self.getSteps(Recipe_ID, "json"):
                step_name = str(row.get("name"))
                step_timer = str(int(row.get("timer")))
                step_temp = str(int(row.get("temp")))
                last_step_temp = step_temp
                sensor = self.kettle.sensor
                if MashIn_Flag == True:
                    if row.get("timer") == 0:
                        step_type = self.mashin if self.mashin != "" else "MashInStep"
                        Notification = "Target temperature reached. Please add malt."
                        MashIn_Flag = False
                        if step_name is None or step_name == "":
                            step_name = "MashIn"
                    elif self.addmashin == "Yes":
                        step_type = self.mashin if self.mashin != "" else "MashInStep"
                        Notification = "Target temperature reached. Please add malt."
                        MashIn_Flag = False
                        step_string = { "name": "MashIn",
                                        "props": {
                                            "AutoMode": self.AutoMode,
                                            "Kettle": self.id,
                                            "Sensor": self.kettle.sensor,
                                            "Temp": self.getJsonMashin(Recipe_ID),
                                            "Timer": 0,
                                            "Notification": Notification
                                            },
                                         "status_text": "",
                                         "status": "I",
                                         "type": step_type
                                        }
                        await self.create_step(step_string)
                        logging.info(step_kettle)  ###################################################

                        step_type = self.mash if self.mash != "" else "MashStep"
                        Notification = ""
                    else:
                        step_type = self.mash if self.mash != "" else "MashStep"
                        Notification = ""

                else:
                    step_type = self.mash if self.mash != "" else "MashStep"
                    Notification = ""

                step_string = { "name": step_name,
                                "props": {
                                        "AutoMode": self.AutoMode,
                                        "Kettle": self.id,
                                        "Sensor": self.kettle.sensor,
                                        "Temp": step_temp,
                                        "Timer": step_timer,
                                        "Notification": Notification
                                        },
                                "status_text": "",
                                "status": "I",
                                "type": step_type
                                }

                await self.create_step(step_string)
            # MashOut -> mashStep to reach mashout-temp for 1 min
            if last_step_temp != e["Abmaischtemperatur"]:
                step_string = { "name": "MashOut",
                                "props": {
                                        "AutoMode": self.AutoMode,
                                        "Kettle": self.id,
                                        "Sensor": self.kettle.sensor,
                                        "Temp": e["Abmaischtemperatur"],
                                        "Timer": 1,
                                        "Notification": ""
                                        },
                                "status_text": "",
                                "status": "I",
                                "type": "MashStep"
                                }

                await self.create_step(step_string)
            # Lautering -> Simple step that sends notification and waits for user input to move to next step (AutoNext=No)
            if self.mashout == "NotificationStep":
                step_string = { "name": "Lautering",
                                "props": {
                                        "AutoNext": "No",
                                        "Kettle": self.id,
                                        "Notification": "Mash Process completed. Please start lautering and press next to start boil."
                                        },
                                    "status_text": "",
                                    "status": "I",
                                    "type": self.mashout
                                    }
                await self.create_step(step_string)
                
            # Measure Original Gravity -> Simple step that sends notification
            step_string = { "name": "Measure Original Gravity",
                            "props": {
                                    "AutoNext": "No",
                                    "Kettle": self.id,
                                    "Notification": "What is the original gravity of the beer wort?"
                                    },
                                "status_text": "",
                                "status": "I",
                                "type": "NotificationStep"
                                }
            await self.create_step(step_string)
                
            # Boil step including hop alarms and alarm for first wort hops -> Automode is set tu yes
            Hops = self.getBoilAlerts(hops, miscs, "json")
            step_kettle = self.boilid
            step_type = self.boil if self.boil != "" else "BoilStep"
            step_time = str(int(boil_time))
            step_temp = self.BoilTemp
            sensor = self.boilkettle.sensor
            LidAlert = "Yes"
            
            logging.info(step_temp)  ###################################################

            step_string = { "name": "Boil Step",
                            "props": {
                                "AutoMode": self.AutoMode,
                                "Kettle": step_kettle,
                                "Sensor": sensor,
                                "Temp": step_temp,
                                "Timer": step_time,
                                "First_Wort": FirstWort,
                                "LidAlert": LidAlert,
                                "Hop_1": Hops[0],
                                "Hop_2": Hops[1],
                                "Hop_3": Hops[2],
                                "Hop_4": Hops[3],
                                "Hop_5": Hops[4],
                                "Hop_6": Hops[5]
                                },
                            "status_text": "",
                            "status": "I",
                            "type": step_type 
                        }

            await self.create_step(step_string)
            
            # Measure Original Gravity -> Simple step that sends notification
            step_string = { "name": "Measure Original Gravity",
                            "props": {
                                    "AutoNext": "No",
                                    "Kettle": self.id,
                                    "Notification": "What is the original gravity of the beer wort?"
                                    },
                                "status_text": "",
                                "status": "I",
                                "type": "NotificationStep"
                                }
            await self.create_step(step_string)

            await self.create_Whirlpool_Cooldown()

            self.cbpi.notify('MMuM-JSON Recipe created ', name, NotificationType.INFO)
        else:
            self.cbpi.notify('Recipe Upload', 'No default Kettle defined. Please specify default Kettle in settings', NotificationType.ERROR)


    async def xml_recipe_creation(self, Recipe_ID):
        config = self.get_config_values()

        if self.kettle is not None:
            # load beerxml file located in upload folder
            self.path = self.cbpi.config_folder.get_upload_file("beer.xml")
            if os.path.exists(self.path) is False:
                self.cbpi.notify("File Not Found", "Please upload a Beer.xml File", NotificationType.ERROR)


            e = xml.etree.ElementTree.parse(self.path).getroot()
            recipe = e.find('./RECIPE[%s]' % (str(Recipe_ID)))
            hops = recipe.findall('./HOPS/HOP')
            miscs = recipe.findall('MISCS/MISC[USE="Boil"]')
            name = e.find('./RECIPE[%s]/NAME' % (str(Recipe_ID))).text
            boil_time = float(e.find('./RECIPE[%s]/BOIL_TIME' % (str(Recipe_ID))).text)
            FirstWort= self.getFirstWort(hops, "xml")

            await self.create_recipe(name)
            # Mash Steps -> first step is different as it heats up to defined temp and stops with notification to add malt
            # AutoMode is yes to start and stop automatic mode or each step
            MashIn_Flag = True
            step_kettle = self.id
            for row in self.getSteps(Recipe_ID, "xml"):
                step_name = str(row.get("name"))
                step_timer = str(int(row.get("timer")))
                step_temp = str(int(row.get("temp")))
                sensor = self.kettle.sensor
                if MashIn_Flag == True:
                    if row.get("timer") == 0:
                        step_type = self.mashin if self.mashin != "" else "MashInStep"
                        Notification = "Target temperature reached. Please add malt."
                        MashIn_Flag = False
                        if step_name is None or step_name == "":
                            step_name = "MashIn"
                    elif self.addmashin == "Yes":
                        step_type = self.mashin if self.mashin != "" else "MashInStep"
                        Notification = "Target temperature reached. Please add malt."
                        MashIn_Flag = False
                        step_string = { "name": "MashIn",
                                        "props": {
                                            "AutoMode": self.AutoMode,
                                            "Kettle": self.id,
                                            "Sensor": self.kettle.sensor,
                                            "Temp": step_temp,
                                            "Timer": 0,
                                            "Notification": Notification
                                            },
                                         "status_text": "",
                                         "status": "I",
                                         "type": step_type
                                        }
                        await self.create_step(step_string)

                        step_type = self.mash if self.mash != "" else "MashStep"
                        Notification = ""
                    else:
                        step_type = self.mash if self.mash != "" else "MashStep"
                        Notification = ""

                else:
                    step_type = self.mash if self.mash != "" else "MashStep"
                    Notification = ""

                step_string = { "name": step_name,
                                "props": {
                                        "AutoMode": self.AutoMode,
                                        "Kettle": self.id,
                                        "Sensor": self.kettle.sensor,
                                        "Temp": step_temp,
                                        "Timer": step_timer,
                                        "Notification": Notification
                                        },
                                "status_text": "",
                                "status": "I",
                                "type": step_type
                                }

                await self.create_step(step_string)

            # MashOut -> Simple step that sends notification and waits for user input to move to next step (AutoNext=No)
            if self.mashout == "NotificationStep":
                step_string = { "name": "Lautering",
                                "props": {
                                        "AutoNext": "No",
                                        "Kettle": self.id,
                                        "Notification": "Mash Process completed. Please start lautering and press next to start boil."
                                        },
                                    "status_text": "",
                                    "status": "I",
                                    "type": self.mashout
                                    }
                await self.create_step(step_string)               
                
            # Boil step including hop alarms and alarm for first wort hops -> Automode is set tu yes
            Hops = self.getBoilAlerts(hops, miscs, "xml")
            step_kettle = self.boilid
            step_type = self.boil if self.boil != "" else "BoilStep"
            step_time = str(int(boil_time))
            step_temp = self.BoilTemp
            sensor = self.boilkettle.sensor
            LidAlert = "Yes"

            step_string = { "name": "Boil Step",
                            "props": {
                                "AutoMode": self.AutoMode,
                                "Kettle": step_kettle,
                                "Sensor": sensor,
                                "Temp": step_temp,
                                "Timer": step_time,
                                "First_Wort": FirstWort,
                                "LidAlert": LidAlert,
                                "Hop_1": Hops[0],
                                "Hop_2": Hops[1],
                                "Hop_3": Hops[2],
                                "Hop_4": Hops[3],
                                "Hop_5": Hops[4],
                                "Hop_6": Hops[5]
                                },
                            "status_text": "",
                            "status": "I",
                            "type": step_type 
                        }

            await self.create_step(step_string)

            await self.create_Whirlpool_Cooldown()

            self.cbpi.notify('BeerXML Recipe created ', name, NotificationType.INFO)
        else:
            self.cbpi.notify('Recipe Upload', 'No default Kettle defined. Please specify default Kettle in settings', NotificationType.ERROR)


   # XML functions to retrieve xml repice parameters (if multiple recipes are stored in one xml file, id could be used)

    def getSteps(self, id, recipe_type):
        steps = []
        if recipe_type == "xml":
            e = xml.etree.ElementTree.parse(self.path).getroot()
            for e in e.findall('./RECIPE[%s]/MASH/MASH_STEPS/MASH_STEP' % (str(id))):
                if self.cbpi.config.get("TEMP_UNIT", "C") == "C":
                    temp = float(e.find("STEP_TEMP").text)
                else:
                    temp = round(9.0 / 5.0 * float(e.find("STEP_TEMP").text) + 32, 2)
                steps.append({"name": e.find("NAME").text, "temp": temp, "timer": float(e.find("STEP_TIME").text)})
        elif recipe_type == "json":
            self.path = self.cbpi.config_folder.get_upload_file("mmum.json")
            e = json.load(open(self.path))
            for idx in range(1,self.findMax("Infusion_Rastzeit%%")):
                if self.cbpi.config.get("TEMP_UNIT", "C") == "C":
                    temp = float(e["Infusion_Rasttemperatur{}".format(idx)])
                else:
                    temp = round(9.0 / 5.0 * float(e["Infusion_Rasttemperatur{}".format(idx)]) + 32, 2)

                steps.append({"name": "Rast {}".format(idx), "temp": temp, "timer": float(e["Infusion_Rastzeit{}".format(idx)])})

        return steps

    async def bf_recipe_creation(self, Recipe_ID):
        config = self.get_config_values()

        if self.kettle is not None:

            brewfather = True
            result=[]
            self.bf_url="https://api.brewfather.app/v1/recipes/" + Recipe_ID
            brewfather_user_id = self.cbpi.config.get("brewfather_user_id", None)
            if brewfather_user_id == "" or brewfather_user_id is None:
                brewfather = False

            brewfather_api_key = self.cbpi.config.get("brewfather_api_key", None)
            if brewfather_api_key == "" or brewfather_api_key is None:
                brewfather = False

            if brewfather == True:
                encodedData = base64.b64encode(bytes(f"{brewfather_user_id}:{brewfather_api_key}", "ISO-8859-1")).decode("ascii")
                headers={"Authorization": "Basic %s" % encodedData}
                bf_recipe = ""

                async with aiohttp.ClientSession(headers=headers) as bf_session:
                    async with bf_session.get(self.bf_url) as r:
                        bf_recipe = await r.json()
                    await bf_session.close()
            
            if bf_recipe !="":
                try:
                    StrikeTemp=bf_recipe['data']['strikeTemp']
                except:
                    StrikeTemp = None
                # BF is sending all Temeprature values in °C. If system is running in F, values need to be converted
                if StrikeTemp is not None and self.TEMP_UNIT != "C":
                    StrikeTemp = round((9.0 / 5.0 *  float(StrikeTemp)+ 32))

                RecipeName = bf_recipe['name']
                BoilTime = bf_recipe['boilTime']
                mash_steps=bf_recipe['mash']['steps']
                hops=bf_recipe['hops']
                try:
                    miscs = bf_recipe['miscs']
                except:
                    miscs = None

                FirstWort = self.getFirstWort(hops, "bf")

                await self.create_recipe(RecipeName)

                # Mash Steps -> first step is different as it heats up to defined temp and stops with notification to add malt
                # AutoMode is yes to start and stop automatic mode or each step
                MashIn_Flag = True
                step_kettle = self.id
                for step in mash_steps:
                    try:
                        step_name = step['name']
                        if step_name == "":
                            step_name = "MashStep"
                    except:
                        step_name = "MashStep"


                    step_timer = str(int(step['stepTime']))

                    if self.TEMP_UNIT == "C":
                        step_temp = str(int(step['stepTemp']))
                    else:
                        step_temp = str(round((9.0 / 5.0 * int(step['stepTemp']) + 32)))

                    sensor = self.kettle.sensor
                    if MashIn_Flag == True: 

                        if int(step_timer) == 0:
                            step_type = self.mashin if self.mashin != "" else "MashInStep"
                            Notification = "Target temperature reached. Please add malt."
                            MashIn_Flag = False

                        elif self.addmashin == "Yes":
                            mashin_temp = str(round(StrikeTemp)) if StrikeTemp is not None else step_temp
                            step_type = self.mashin if self.mashin != "" else "MashInStep"
                            Notification = "Target temperature reached. Please add malt."
                            MashIn_Flag = False
                            step_string = { "name": "MashIn",
                                        "props": {
                                            "AutoMode": self.AutoMode,
                                            "Kettle": self.id,
                                            "Sensor": self.kettle.sensor,
                                            "Temp": mashin_temp,
                                            "Timer": 0,
                                            "Notification": Notification
                                            },
                                         "status_text": "",
                                         "status": "I",
                                         "type": step_type
                                        }
                            await self.create_step(step_string)

                            step_type = self.mash if self.mash != "" else "MashStep"
                            Notification = ""
                        else:
                            step_type = self.mash if self.mash != "" else "MashStep"
                            Notification = ""

                    else:
                        step_type = self.mash if self.mash != "" else "MashStep"
                        Notification = ""

                    step_string = { "name": step_name,
                                    "props": {
                                        "AutoMode": self.AutoMode,
                                        "Kettle": self.id,
                                        "Sensor": self.kettle.sensor,
                                        "Temp": step_temp,
                                        "Timer": step_timer,
                                        "Notification": Notification
                                        },
                                     "status_text": "",
                                     "status": "I",
                                     "type": step_type
                                    }

                    await self.create_step(step_string)

                # MashOut -> Simple step that sends notification and waits for user input to move to next step (AutoNext=No)

                if self.mashout == "NotificationStep":
                    step_string = { "name": "Lautering",
                                    "props": {
                                        "AutoNext": "No",
                                        "Kettle": self.id,
                                        "Notification": "Mash Process completed. Please start lautering and press next to start boil."
                                        },
                                    "status_text": "",
                                    "status": "I",
                                    "type": self.mashout
                                    }
                await self.create_step(step_string)    

                # Boil step including hop alarms and alarm for first wort hops -> Automode is set tu yes
                Hops = self.getBoilAlerts(hops , miscs, "bf")

                step_kettle = self.boilid
                step_time = str(int(BoilTime))
                step_type = self.boil if self.boil != "" else "BoilStep"
                step_temp = self.BoilTemp
                sensor = self.boilkettle.sensor
                LidAlert = "Yes"

                step_string = { "name": "Boil Step",
                            "props": {
                                "AutoMode": self.AutoMode,
                                "Kettle": step_kettle,
                                "Sensor": sensor,
                                "Temp": step_temp,
                                "Timer": step_time,
                                "First_Wort": FirstWort,
                                "LidAlert": LidAlert,
                                "Hop_1": Hops[0],
                                "Hop_2": Hops[1],
                                "Hop_3": Hops[2],
                                "Hop_4": Hops[3],
                                "Hop_5": Hops[4],
                                "Hop_6": Hops[5]
                                },
                            "status_text": "",
                            "status": "I",
                            "type": step_type 
                        }

                await self.create_step(step_string)

                await self.create_Whirlpool_Cooldown()

                self.cbpi.notify('Brewfather App Recipe created: ', RecipeName, NotificationType.INFO)
        else:
            self.cbpi.notify('Recipe Upload', 'No default Kettle defined. Please specify default Kettle in settings', NotificationType.ERROR)

    def getBoilAlerts(self, hops, miscs, recipe_type):
        alerts = []
        for hop in hops:
            if recipe_type == "xml":
                use = hop.find('USE').text
                ## Hops which are not used in the boil step should not cause alerts
                if use != 'Aroma' and use != 'Boil':
                    continue
                alerts.append(float(hop.find('TIME').text))
            elif recipe_type == "bf":
                use = hop['use']
                if use != 'Aroma' and use != 'Boil':
                    continue
                alerts.append(float(hop['time']))
            elif recipe_type == "kbh":
                alerts.append(float(hop[0]))
            elif  recipe_type == "json":
                alerts.append(float(hop['time']))
                
        ## There might also be miscelaneous additions during boild time
        if miscs is not None:
            for misc in miscs:
                if recipe_type == "xml":
                    alerts.append(float(misc.find('TIME').text))
                elif recipe_type == "bf":
                    use = misc['use']
                    if use != 'Aroma' and use != 'Boil':
                        continue
                    alerts.append(float(misc['time']))
                elif recipe_type == "kbh":
                    alerts.append(float(misc[0]))
                elif  recipe_type == "json":
                    alerts.append(float(misc['time']))
        ## Dedupe and order the additions by their time, to prevent multiple alerts at the same time
        alerts = sorted(list(set(alerts)))
        ## CBP should have these additions in reverse
        alerts.reverse()
        hop_alerts = []
        for i in range(0,6):
            try:
                hop_alerts.append(str(int(alerts[i])))
            except:
                hop_alerts.append(None)
        return hop_alerts

    def getFirstWort(self, hops, recipe_type):
        alert = "No"
        if recipe_type == "kbh":
            if len(hops) != 0:
                alert = "Yes"
        elif recipe_type == "xml":
            for hop in hops:
                use = hop.find('USE').text
                ## Hops which are not used in the boil step should not cause alerts
                if use != 'First Wort':
                    continue
                alert = "Yes"
        elif recipe_type == "bf":
            for hop in hops:
                if hop['use'] == "First Wort":
                    alert="Yes"
        elif recipe_type == "json":
            for hop in hops:
                alert="Yes"
        return alert

    async def create_Whirlpool_Cooldown(self):
        # Add Waitstep as Whirlpool
        if self.cooldown != "WaiStep" and self.cooldown !="":
            step_string = { "name": "Whirlpool",
                            "props": {
                                "Kettle": self.boilid,
                                "Timer": "15"
                                },
                            "status_text": "",
                            "status": "I",
                            "type": "WaitStep" 
                        }
            await self.create_step(step_string)

        # CoolDown step is sending a notification when cooldowntemp is reached
        step_type = self.cooldown if self.cooldown != "" else "WaitStep"
        step_name = "CoolDown"
        cooldown_sensor = ""
        step_temp = ""
        step_timer = "15"
        if step_type == "CooldownStep":
            cooldown_sensor = self.cbpi.config.get("steps_cooldown_sensor", None)
            if cooldown_sensor is None or cooldown_sensor == '':
                cooldown_sensor = self.boilkettle.sensor  # fall back to boilkettle sensor if no other sensor is specified
            step_timer = ""                
            step_temp = int(self.CoolDownTemp)
            step_string = { "name": "Cooldown",
                            "props": {
                                "Kettle": self.boilid,
                                "Timer": step_timer,
                                "Temp": step_temp,
                                "Sensor": cooldown_sensor,
                                "Actor": self.CoolDownActor
                                },
                            "status_text": "",
                            "status": "I",
                            "type": step_type 
                        }
            await self.create_step(step_string)

    def get_config_values(self):
        self.kettle = None
        self.boilkettle = None
        #Define MashSteps
        self.TEMP_UNIT = self.cbpi.config.get("TEMP_UNIT", "C")
        self.AutoMode = self.cbpi.config.get("AutoMode", "Yes")
        self.mashin =  self.cbpi.config.get("steps_mashin", "MashInStep")
        self.mash = self.cbpi.config.get("steps_mash", "MashStep") 
        self.mashout = self.cbpi.config.get("steps_mashout", None) # Currently used only for the Braumeister 
        self.boil = self.cbpi.config.get("steps_boil", "BoilStep") 
        self.whirlpool="Waitstep"
        self.cooldown = self.cbpi.config.get("steps_cooldown", "WaitStep") 
        #get default boil temp from settings
        self.BoilTemp = self.cbpi.config.get("steps_boil_temp", 98)
        #get default cooldown temp alarm setting
        self.CoolDownTemp = self.cbpi.config.get("steps_cooldown_temp", 25)
        self.CoolDownActor = self.cbpi.config.get("steps_cooldown_actor", None)
        # get default Kettle from Settings       
        self.id = self.cbpi.config.get('MASH_TUN', None)
        self.boilid = self.cbpi.config.get('BoilKettle', None)
        if self.boilid is None:
            self.boilid = self.id
        # If next parameter is Yes, MashIn Ste will be added before first mash step if not included in recipe
        self.addmashin = self.cbpi.config.get('AddMashInStep', "Yes")

        try:
            self.kettle = self.cbpi.kettle.find_by_id(self.id) 
        except:
            self.cbpi.notify('Recipe Upload', 'No default Kettle defined. Please specify default Kettle in settings', NotificationType.ERROR)
        try:
            self.boilkettle = self.cbpi.kettle.find_by_id(self.boilid) 
        except:
            self.boilkettle = self.kettle

        config_values = { "kettle": self.kettle,
                          "kettle_id": str(self.id),
                          "boilkettle": self.boilkettle,
                          "boilkettle_id": str(self.boilid),
                          "mashin": str(self.mashin),
                          "mash": str(self.mash),
                          "mashout": str(self.mashout),
                          "boil": str(self.boil),
                          "whirlpool": str(self.whirlpool),
                          "cooldown": str(self.cooldown),
                          "boiltemp": str(self.BoilTemp),
                          "cooldowntemp": str(self.CoolDownTemp),
                          "cooldownactor": self.CoolDownActor,
                          "temp_unit": str(self.TEMP_UNIT),
                          "AutoMode": str(self.AutoMode)
                        }
        logging.info(config_values)
        return config_values

    async def create_recipe(self, name):
        # Create recipe in recipe Book with name of first recipe in xml file
        self.recipeID = await self.cbpi.recipe.create(name)
        # send recipe to mash profile
        await self.cbpi.recipe.brew(self.recipeID)
        # remove empty recipe from recipe book
        await self.cbpi.recipe.remove(self.recipeID)

    # function to create json to be send to api to add a step to the current mash profile. Currently all properties are send to each step which does not cuase an issue
    async def create_step(self, step_string):
        #get server port from settings and define url for api calls -> adding steps
        self.port = str(self.cbpi.static_config.get('port',8000))
        self.url="http://127.0.0.1:" + self.port + "/step2/"
        # convert step:string to json required for api call. 
        step = json.dumps(step_string)
        headers={'Content-Type': 'application/json', 'Accept': 'application/json'}
        async with aiohttp.ClientSession(headers=headers) as session:
            async with session.post(self.url, data=step) as response:
                return await response.text()
            await self.push_update()



import logging
import asyncio
from cbpi.api import *
from cbpi.api.config import ConfigType
from cbpi.api.base import CBPiBase
from cbpi.controller.fermentation_controller import FermentationController
from cbpi.controller.kettle_controller import KettleController

logger = logging.getLogger(__name__)

class MQTTUtil(CBPiExtension):

    def __init__(self,cbpi):
        self.cbpi = cbpi
        self.kettlecontroller = cbpi.kettle
        self.fermentationcontroller = cbpi.fermenter
        # sensor and actor update is done anyhow during startup
#        self.sensorcontroller = cbpi.sensor
#        self.actorcontroller = cbpi.actor

        self.mqttupdate = int(self.cbpi.config.get("MQTTUpdate", 0))
        if self.mqttupdate != 0:
            self._task = asyncio.create_task(self.run())
            logger.info("INIT MQTTUtil")
        else:
            self._task = asyncio.create_task(self.push_once())

    async def push_once(self):
        # wait some time to ensure that kettlecontroller is started
        await asyncio.sleep(5)
        self.push_update()


    async def run(self):

        while True:
            self.push_update()
            await asyncio.sleep(self.mqttupdate)

    def push_update(self):
#        try:
#            self.actor=self.actorcontroller.get_state()
#            for item in self.actor['data']:
#                self.cbpi.push_update("cbpi/{}/{}".format("actorupdate",item['id']), item)
#        except Exception as e:
#            logging.error(e)
#            pass
#        try:
#            self.sensor=self.sensorcontroller.get_state()
#            for item in self.sensor['data']:
#                self.cbpi.push_update("cbpi/{}/{}".format("sensorupdate",item['id']), item)
#        except Exception as e:
#            logging.error(e)
#            pass
        try:
            self.kettle=self.kettlecontroller.get_state()
            for item in self.kettle['data']:
                self.cbpi.push_update("cbpi/{}/{}".format("kettleupdate",item['id']), item)
        except Exception as e:
            logging.error(e)
            pass
        try:
            self.fermenter=self.fermentationcontroller.get_state()
            for item in self.fermenter['data']:
                self.cbpi.push_update("cbpi/{}/{}".format("fermenterupdate",item['id']), item)
        except Exception as e:
            logging.error(e)
            pass


def setup(cbpi):
    if str(cbpi.static_config.get("mqtt", False)).lower() == "true":
        cbpi.plugin.register("MQTTUtil", MQTTUtil)
    pass



__all__ = ["CBPiException","KettleException","FermenterException","SensorException","ActorException"]


class CBPiException(Exception):
    pass

class KettleException(CBPiException):
    pass

class FermenterException(CBPiException):
    pass

class SensorException(CBPiException):
    pass

class ActorException(CBPiException):
    pass




import importlib
import logging
import os
import pkgutil

from importlib import import_module

from cbpi.api import *
from cbpi.utils.utils import load_config
from importlib_metadata import metadata, version

logger = logging.getLogger(__name__)

class PluginController():
    modules = {}
    types = {}

    def __init__(self, cbpi):
        self.cbpi = cbpi

    def load_plugins(self):

        this_directory = os.sep.join(
            os.path.abspath(__file__).split(os.sep)[:-1])
        for filename in os.listdir(os.path.join(this_directory, "../extension")):
            if os.path.isdir(
                    os.path.join(this_directory, "../extension/") + filename) is False or filename == "__pycache__":
                continue
            try:
                logger.info("Trying to load plugin %s" % filename)
                data = load_config(os.path.join(
                    this_directory, "../extension/%s/config.yaml" % filename))
                if (data.get("active") is True and data.get("version") == 4):
                    self.modules[filename] = import_module(
                        "cbpi.extension.%s" % (filename))
                    self.modules[filename].setup(self.cbpi)
                    # logger.info("Plugin %s loaded successful" % filename)
                else:
                    logger.warning(
                        "Plugin %s is not supporting version 4" % filename)

            except Exception as e:

                logger.error(e)

    def load_plugins_from_evn(self):

        discovered_plugins = {
            name: importlib.import_module(name)
            for finder, name, ispkg
            in pkgutil.iter_modules()
            if name.startswith('cbpi') and len(name) > 4
        }

        for key, value in discovered_plugins.items():
            from importlib.metadata import version
            try:
                logger.info("Try to load plugin:  {} == {} ".format(
                    key, version(key)))
                value.setup(self.cbpi)
                logger.info("Plugin {} loaded successfully".format(key))
            except Exception as e:
                logger.error("FAILED to load plugin {} ".format(key))
                logger.error(e)

    def register(self, name, clazz) -> None:
        '''
        Register a new actor type
        :param name: actor name
        :param clazz: actor class
        :return: None
        '''
        logger.debug("Register %s Class %s" % (name, clazz.__name__))

        if issubclass(clazz, CBPiActor):
            self.cbpi.actor.types[name] = self._parse_step_props(clazz, name)

        if issubclass(clazz, CBPiKettleLogic):
            self.cbpi.kettle.types[name] = self._parse_step_props(clazz, name)

        if issubclass(clazz, CBPiFermenterLogic):
            self.cbpi.fermenter.types[name] = self._parse_step_props(
                clazz, name)

        if issubclass(clazz, CBPiSensor):
            self.cbpi.sensor.types[name] = self._parse_step_props(clazz, name)

        if issubclass(clazz, CBPiStep):
            self.cbpi.step.types[name] = self._parse_step_props(clazz, name)

        if issubclass(clazz, CBPiFermentationStep):
            self.cbpi.fermenter.steptypes[name] = self._parse_step_props(
                clazz, name)

        if issubclass(clazz, CBPiExtension):
            self.c = clazz(self.cbpi)

    def _parse_property_object(self, p):
        if isinstance(p, Property.Number):
            return {"label": p.label, "type": "number", "configurable": p.configurable, "description": p.description,
                    "default_value": p.default_value}
        elif isinstance(p, Property.Text):
            return {"label": p.label, "type": "text", "configurable": p.configurable, "default_value": p.default_value,
                    "description": p.description}
        elif isinstance(p, Property.Select):
            return {"label": p.label, "type": "select", "configurable": True, "options": p.options,
                    "description": p.description}
        elif isinstance(p, Property.Actor):
            return {"label": p.label, "type": "actor", "configurable": p.configurable, "description": p.description}
        elif isinstance(p, Property.Sensor):
            return {"label": p.label, "type": "sensor", "configurable": p.configurable, "description": p.description}
        elif isinstance(p, Property.Kettle):
            return {"label": p.label, "type": "kettle", "configurable": p.configurable, "description": p.description}
        elif isinstance(p, Property.Fermenter):
            return {"label": p.label, "type": "fermenter", "configurable": p.configurable, "description": p.description}

    def _parse_step_props(self, cls, name):

        result = {"name": name, "class": cls, "properties": [], "actions": []}

        if hasattr(cls, "cbpi_parameters"):
            parameters = []
            for p in cls.cbpi_parameters:
                parameters.append(self._parse_property_object(p))
            result["properties"] = parameters
        for method_name, method in cls.__dict__.items():
            if hasattr(method, "action"):
                key = method.__getattribute__("key")
                parameters = []
                for p in method.__getattribute__("parameters"):
                    parameters.append(self._parse_property_object(p))
                result["actions"].append(
                    {"method": method_name, "label": key, "parameters": parameters})

        return result

    def _parse_props(self, cls):

        name = cls.__name__

        result = {"name": name, "class": cls, "properties": [], "actions": []}

        tmpObj = cls(cbpi=None, managed_fields=None)
        members = [attr for attr in dir(tmpObj) if not callable(
            getattr(tmpObj, attr)) and not attr.startswith("__")]
        for m in members:
            if isinstance(tmpObj.__getattribute__(m), Property.Number):
                t = tmpObj.__getattribute__(m)
                result["properties"].append(
                    {"name": m, "label": t.label, "type": "number", "configurable": t.configurable,
                     "description": t.description, "default_value": t.default_value})
            elif isinstance(tmpObj.__getattribute__(m), Property.Text):
                t = tmpObj.__getattribute__(m)
                result["properties"].append(
                    {"name": m, "label": t.label, "type": "text", "configurable": t.configurable,
                     "default_value": t.default_value, "description": t.description})
            elif isinstance(tmpObj.__getattribute__(m), Property.Select):
                t = tmpObj.__getattribute__(m)
                result["properties"].append(
                    {"name": m, "label": t.label, "type": "select", "configurable": True, "options": t.options,
                     "description": t.description})
            elif isinstance(tmpObj.__getattribute__(m), Property.Actor):
                t = tmpObj.__getattribute__(m)
                result["properties"].append(
                    {"name": m, "label": t.label, "type": "actor", "configurable": t.configurable,
                     "description": t.description})
            elif isinstance(tmpObj.__getattribute__(m), Property.Sensor):
                t = tmpObj.__getattribute__(m)
                result["properties"].append(
                    {"name": m, "label": t.label, "type": "sensor", "configurable": t.configurable,
                     "description": t.description})
            elif isinstance(tmpObj.__getattribute__(m), Property.Kettle):
                t = tmpObj.__getattribute__(m)
                result["properties"].append(
                    {"name": m, "label": t.label, "type": "kettle", "configurable": t.configurable,
                     "description": t.description})
            elif isinstance(tmpObj.__getattribute__(m), Property.Fermenter):
                t = tmpObj.__getattribute__(m)
                result["properties"].append(
                    {"name": m, "label": t.label, "type": "fermenter", "configurable": t.configurable,
                     "description": t.description})

        for method_name, method in cls.__dict__.items():
            if hasattr(method, "action"):
                key = method.__getattribute__("key")
                parameters = method.__getattribute__("parameters")
                result["actions"].append(
                    {"method": method_name, "label": key, "parameters": parameters})

        return result

    async def load_plugin_list(self):
        result = []
        try:
            discovered_plugins = {
                name: importlib.import_module(name)
                for finder, name, ispkg
                in pkgutil.iter_modules()
                if name.startswith('cbpi') and len(name) > 4
            }
            for key, module in discovered_plugins.items():
                from importlib.metadata import version
                try:
                    from importlib.metadata import (distribution, metadata,
                                                    version)
                    meta = metadata(key)
                    result.append({row: meta[row]
                                  for row in list(metadata(key))})
                except Exception as e:
                    logger.error("FAILED to load plugin {} ".format(key))
                    logger.error(e)

        except Exception as e:
            logger.error(e)
            return []
        return result



from aiohttp.test_utils import unittest_run_loop
from tests.cbpi_config_fixture import CraftBeerPiTestCase

from cbpi.craftbeerpi import CraftBeerPi


class DashboardTestCase(CraftBeerPiTestCase):

    @unittest_run_loop
    async def test_crud(self):
        data = {
            "name": "MyDashboard",

        }

        dashboard_content = {
            "type": "Test",
            "x": 0,
            "y": 0,
            "config": {}
        }

        resp = await self.client.get(path="/dashboard/current")
        assert resp.status == 200

        dashboard_id = await resp.json()

         # Add dashboard content
        dashboard_content["dbid"] = dashboard_id
        resp = await self.client.post(path="/dashboard/%s/content" % dashboard_id, json=dashboard_content)
        assert resp.status == 204


# content of conftest.py
from codecs import ignore_errors
from distutils.command.config import config
import os
from cbpi.configFolder import ConfigFolder
from cbpi.craftbeerpi import CraftBeerPi
from aiohttp.test_utils import AioHTTPTestCase
from distutils.dir_util import copy_tree


class CraftBeerPiTestCase(AioHTTPTestCase):

    async def get_application(self):
        self.config_folder = self.configuration()
        self.cbpi = CraftBeerPi(self.config_folder)
        await self.cbpi.init_serivces()
        return self.cbpi.app

    def configuration(self):
        test_directory = os.path.dirname(__file__)
        test_config_directory = os.path.join(test_directory, 'cbpi-test-config')
        configFolder = ConfigFolder(test_config_directory)
        return configFolder



from aiohttp.test_utils import unittest_run_loop
from tests.cbpi_config_fixture import CraftBeerPiTestCase


class IndexTestCase(CraftBeerPiTestCase):

    @unittest_run_loop
    async def test_endpoints(self):
        # Test Index Page
        resp = await self.client.post(path="/system/restart")
        assert resp.status == 200

        resp = await self.client.post(path="/system/shutdown")
        assert resp.status == 200




import time
import asyncio
import math


class Timer(object):

    def __init__(self, timeout, on_done = None, on_update = None) -> None:
        super().__init__()
        self.timeout = timeout
        self._timemout = self.timeout
        self._task = None
        self._callback = on_done
        self._update = on_update
        self.start_time = None
        self.end_time = None
    
    def done(self, task):
        if self._callback is not None:
            asyncio.create_task(self._callback(self))

    async def _job(self):
        self.start_time = int(time.time())
        self.end_time = self.start_time + int(round(self._timemout, 0))
        self.count = self.end_time - self.start_time
        try:
            while self.count > 0:
                self.count = (self.end_time - int(time.time()))
                if self._update is not None:
                    await self._update(self,self.count)
                await asyncio.sleep(1)
        except asyncio.CancelledError:
            end = int(time.time())
            duration = end - self.start_time
            self._timemout = self._timemout - duration

    async def add(self, seconds):
        self.end_time = self.end_time + seconds

    def start(self):
        self._task = asyncio.create_task(self._job())
        self._task.add_done_callback(self.done)

    async def stop(self):
        if self._task and self._task.done() is False:
            self._task.cancel()
            await self._task

    def reset(self):
        if self.is_running is True:
            return
        self._timemout = self.timeout

    def is_running(self):
        return not self._task.done()

    def set_time(self,timeout):
        if self.is_running is True:
            return
        self.timeout = timeout

    def get_time(self):
        return self.format_time(int(round(self._timemout,0)))

    @classmethod
    def format_time(cls, time):
            pattern_h = '{0:02d}:{1:02d}:{2:02d}'
            pattern_d = '{0:02d}D {1:02d}:{2:02d}:{3:02d}'
            seconds = time % 60
            minutes = math.floor(time / 60) % 60
            hours = math.floor(time / 3600) % 24
            days = math.floor(time / 86400)
            if days != 0:
                remaining_time = pattern_d.format(days, hours, minutes, seconds)
            else:
                remaining_time = pattern_h.format(hours, minutes, seconds)
            return remaining_time



import asyncio
from asyncio import tasks
import logging
from cbpi.api import *

@parameters([Property.Number(label="OffsetOn", configurable=True, description="Offset below target temp when heater should switched on"),
             Property.Number(label="OffsetOff", configurable=True, description="Offset below target temp when heater should switched off")])
class Hysteresis(CBPiKettleLogic):
    
    async def run(self):
        try:
            self.offset_on = float(self.props.get("OffsetOn", 0))
            self.offset_off = float(self.props.get("OffsetOff", 0))
            self.kettle = self.get_kettle(self.id)
            self.heater = self.kettle.heater
            logging.info("Hysteresis {} {} {} {}".format(self.offset_on, self.offset_off, self.id, self.heater))

            # self.get_actor_state()
	    

            while self.running == True:
                
                sensor_value = self.get_sensor_value(self.kettle.sensor).get("value")
                target_temp = self.get_kettle_target_temp(self.id)
                if sensor_value < target_temp - self.offset_on:
                    await self.actor_on(self.heater)
                elif sensor_value >= target_temp - self.offset_off:
                    await self.actor_off(self.heater)
                await asyncio.sleep(1)

        except asyncio.CancelledError as e:
            pass
        except Exception as e:
            logging.error("CustomLogic Error {}".format(e))
        finally:
            self.running = False
            await self.actor_off(self.heater)



def setup(cbpi):

    '''
    This method is called by the server during startup 
    Here you need to register your plugins at the server
    
    :param cbpi: the cbpi core 
    :return: 
    '''

    cbpi.plugin.register("Hysteresis", Hysteresis)



"""Jobs scheduler for managing background task (asyncio).
The library gives controlled way for scheduling background tasks for
asyncio applications.
"""


__version__ = '0.2.2'

import asyncio
from ._scheduler import Scheduler


async def create_scheduler(cbpi,*, close_timeout=0.1, limit=100,
                           pending_limit=10000, exception_handler=None):
    if exception_handler is not None and not callable(exception_handler):
        raise TypeError('A callable object or None is expected, '
                        'got {!r}'.format(exception_handler))
    loop = asyncio.get_event_loop()
    return Scheduler(cbpi=cbpi,loop=loop, close_timeout=close_timeout,
                     limit=limit, pending_limit=pending_limit,
                     exception_handler=exception_handler)

__all__ = ('create_scheduler',)


from aiohttp import web
from cbpi.api import request_mapping
from cbpi.utils import json_dumps
import logging


class PluginHttpEndpoints:

    def __init__(self,cbpi):
        self.cbpi = cbpi
        self.cbpi.register(self, url_prefix="/plugin")

    @request_mapping(path="/install/", method="POST", auth_required=False, json_schema={"package_name": str})
    async def install(self, request):
        """
        ---
        description: Install Plugin
        tags:
        - Plugin
        parameters:
        - in: body
          name: body
          description: Install a plugin
          required: true
          schema:
            type: object
            properties:
              package_name:
                type: string
        produces:
        - application/json
        responses:
            "204":
                description: successful operation. Return "pong" text
            "405":
                description: invalid HTTP Method
        """

        data = await request.json()
        return web.Response(status=204) if await self.cbpi.plugin.install(data["package_name"]) is True else web.Response(status=500)

    @request_mapping(path="/uninstall", method="POST", auth_required=False, json_schema={"package_name": str})
    async def uninstall(self, request):
        """
        ---
        description: Uninstall Plugin
        tags:
        - Plugin
        parameters:
        - in: body
          name: body
          description: Uninstall a plugin
          required: true
          schema:
            type: object
            properties:
              package_name:
                type: string
        produces:
        - application/json
        responses:
            "204":
                description: successful operation. Return "pong" text
            "405":
                description: invalid HTTP Method
        """

        data = await request.json()
        return web.Response(status=204) if await self.cbpi.plugin.uninstall(data["package_name"]) is True else web.Response(status=500)


    @request_mapping(path="/list", method="GET", auth_required=False)
    async def list(self, request):
        """
            ---
            description: Get a list of avialable plugins
            tags:
            - Plugin
            produces:
            - application/json
            responses:
                "200":
                    description: successful operation. Return "pong" text
                "405":
                    description: invalid HTTP Method
            """
        plugin_list = await self.cbpi.plugin.load_plugin_list()
        return web.json_response(plugin_list, dumps=json_dumps)



# -*- coding: utf-8 -*-
import asyncio
from aiohttp import web
from cbpi.api import *

import re
import random


cache = {}

@parameters([Property.Text(label="Key", configurable=True, description="Http Key")])
class HTTPSensor(CBPiSensor):
    def __init__(self, cbpi, id, props):
        super(HTTPSensor, self).__init__(cbpi, id, props)
        self.running = True
        self.value = 0

    async def run(self):
        '''
        This method is executed asynchronousely 
        In this example the code is executed every second
        '''
        while self.running is True:
            try:
                cache_value = cache.pop(self.props.get("Key"), None)
                if cache_value is not None:
                    self.value = float(cache_value)
                    self.push_update(self.value)
            except Exception as e:
                pass
            await asyncio.sleep(1)

    def get_state(self):
        # return the current state of the sensor
        return dict(value=self.value)

class HTTPSensorEndpoint(CBPiExtension):


    def __init__(self, cbpi):
        '''
        Initializer

        :param cbpi:
        '''
        self.pattern_check = re.compile("^[a-zA-Z0-9,.]{0,10}$")

        self.cbpi = cbpi
        # register component for http, events
        # In addtion the sub folder static is exposed to access static content via http
        self.cbpi.register(self, "/httpsensor")


    @request_mapping(path="/{key}/{value}", auth_required=False)
    async def http_new_value2(self, request):
        """
        ---
        description: Kettle Heater on
        tags:
        - HttpSensor
        parameters:
        - name: "key"
          in: "path"
          description: "Sensor Key"
          required: true
          type: "string"
        - name: "value"
          in: "path"
          description: "Value"
          required: true
          type: "integer"
          format: "int64"
        responses:
            "204":
                description: successful operation
        """

        global cache
        key = request.match_info['key']
        value = request.match_info['value']
        if self.pattern_check.match(key) is None:
            return web.json_response(status=422, data={'error': "Key not matching pattern ^[a-zA-Z0-9,.]{0,10}$"})

        if self.pattern_check.match(value) is None:
            return web.json_response(status=422, data={'error': "Data not matching pattern ^[a-zA-Z0-9,.]{0,10}$"})

        
        cache[key] = value

        return web.Response(status=204)


def setup(cbpi):

    '''
    This method is called by the server during startup 
    Here you need to register your plugins at the server
    
    :param cbpi: the cbpi core 
    :return: 
    '''
    cbpi.plugin.register("HTTPSensor", HTTPSensor)
    cbpi.plugin.register("HTTPSensorEndpoint", HTTPSensorEndpoint)




import asyncio
import inspect
import logging

from cbpi.api import *


class CBPiEventBus(object):


    class Node(object):
        __slots__ = '_children', '_content'

        def __init__(self):
            self._children = {}
            self._content = None

    class Content(object):
        def __init__(self, parent, topic, method, once, supports_future=False):
            self.parent = parent
            self.method = method
            self.name = method.__name__
            self.once = once
            self.topic = topic
            self.supports_future = supports_future

    class Result:

        def __init__(self, result, timeout):
            self.result = result
            self.timeout = timeout

    class ResultContainer:

        def __init__(self, results, timeout=False):
            self.results = {}
            self.timeout = timeout
            self._jobs = set()
            for key, value in results.items():
                if value.done() is True:
                    self.results[key] = CBPiEventBus.Result(value.result(), True)
                else:
                    self.results[key] = CBPiEventBus.Result(None, False)

        def get(self, key):
            r = self.results.get(key)
            if r is None:
                raise CBPiException("Event Key %s not found." % key)
            return (r.result, r.timeout)


    def register(self, topic, method, once=False):

        if method in self.registry:
            raise RuntimeError("Method %s already registerd. Please unregister first!" % method.__name__)
        self.logger.info("Topic %s", topic)

        node = self._root
        for sym in topic.split('/'):
            node = node._children.setdefault(sym, self.Node())

        if not isinstance(node._content, list):
            node._content = []

        sig = inspect.signature(method)


        if "future" in sig.parameters:
            supports_future = True
        else:
            supports_future = False
        c = self.Content(node, topic, method, once, supports_future)
        node._content.append(c)
        self.registry[method] = c

    def get_callbacks(self, key):
        try:
            node = self._root
            for sym in key.split('/'):
                node = node._children[sym]
            if node._content is None:
                raise KeyError(key)
            return node._content
        except KeyError:
            raise KeyError(key)

    def unregister(self, method):
        self.logger.info("Unregister %s", method.__name__)
        if method in self.registry:
            content = self.registry[method]
            clean_idx = None
            for idx, content_obj in enumerate(content.parent._content):
                if method == content_obj.method:
                    clean_idx = idx
                    break
            if clean_idx is not None:
                del content.parent._content[clean_idx]

    def __init__(self, loop, cbpi):
        self.logger = logging.getLogger(__name__)
        self.cbpi = cbpi
        self._root = self.Node()
        self.registry = {}
        self.docs = {}
        if loop is not None:
            self.loop = loop
        else:
            self.loop = asyncio.get_event_loop()


    def sync_fire(self,topic: str,timeout=1, **kwargs):
        self.loop.create_task(self.fire(topic=topic, timeout=timeout, **kwargs))

    async def fire(self, topic: str, timeout=0.5, **kwargs):

        futures = {}


        #self.logger.info("FIRE %s %s" % (topic, kwargs))

        async def wait(futures):
            if(len(futures) > 0):
                await asyncio.wait(futures.values())


        for e in self.iter_match(topic):
            content_array = e
            keep_idx = []
            for idx, content_obj in enumerate(content_array):

                if inspect.iscoroutinefunction(content_obj.method):
                    if content_obj.supports_future is True:
                        fut = self.loop.create_future()
                        futures["%s.%s" % (content_obj.method.__module__, content_obj.name)] = fut
                        self.loop.create_task(content_obj.method(**kwargs, topic = topic, future=fut))

                    else:
                        self.loop.create_task(content_obj.method(**kwargs, topic=topic))
                else:
                    # only asnyc
                    pass
                if content_obj.once is False:
                    keep_idx.append(idx)

            # FILTER only elements with are required
            if len(keep_idx) < len(e):
                e[0].parent._content = [e[0].parent._content[i] for i in keep_idx]

        if timeout is not None:
            try:
                await asyncio.wait_for(wait(futures), timeout=timeout)
                is_timedout = False
            except asyncio.TimeoutError:
                is_timedout = True
            return self.ResultContainer(futures, is_timedout)


    def dump(self):
        def rec(node, i=0):
            result = []
            if node._content is not None:
                for c in node._content:

                    result.append(dict(topic=c.topic, supports_future=c.supports_future, method=c.method.__name__, path=c.method.__module__, once=c.once))

            if node._children is not None:
                for c in node._children:
                    result = result + rec(node._children[c], i + 1)
            return result

        result = rec(self._root)

        return result

    def iter_match(self, topic):

        lst = topic.split('/')
        normal = not topic.startswith('$')

        def rec(node, i=0):
            if i == len(lst):
                if node._content is not None:
                    yield node._content
            else:
                part = lst[i]
                if part in node._children:
                    for content in rec(node._children[part], i + 1):
                        yield content
                if '+' in node._children and (normal or i > 0):
                    for content in rec(node._children['+'], i + 1):
                        yield content
            if '#' in node._children and (normal or i > 0):
                content = node._children['#']._content
                if content is not None:
                    yield content

        return rec(self._root)

    def register_object(self, obj):

        for method in [getattr(obj, f) for f in dir(obj) if callable(getattr(obj, f)) and hasattr(getattr(obj, f), "eventbus")]:
            doc = None
            if method.__doc__ is not None:
                try:
                    doc = yaml.load(method.__doc__)
                    doc["topic"] = method.__getattribute__("topic")
                except:
                    pass
            self.register(method.__getattribute__("topic"), method)


import logging
import os
import sys

import yaml

__all__ = ["CBPiExtension"]



logger = logging.getLogger(__file__)



class CBPiExtension():

    def init(self):
        pass

    def stop(self):
        pass

    def __init__(self, *args, **kwds):

        for a in kwds:
            logger.debug("Parameter: %s Value: %s" % ( a, kwds.get(a)))
            super(CBPiExtension, self).__setattr__(a, kwds.get(a))
        self.cbpi = kwds.get("cbpi")
        self.id = kwds.get("id")
        self.value = None
        self.__dirty = False

    def __setattr__(self, name, value):

        if name != "_CBPiExtension__dirty":
            self.__dirty = True
            super(CBPiExtension, self).__setattr__(name, value)
        else:
            super(CBPiExtension, self).__setattr__(name, value)

    def load_config(self):

        path = os.path.dirname(sys.modules[self.__class__.__module__].__file__)

        try:
            with open("%s/config.yaml" % path, 'rt') as f:
                data = yaml.load(f)

            return data
        except:
            logger.warning("Failed to load config %s/config.yaml" % path)

    




import logging
import requests
from cbpi.configFolder import ConfigFolder
from cbpi.utils.utils import load_config
from zipfile import ZipFile
from cbpi.craftbeerpi import CraftBeerPi
import os
import pkgutil
import shutil
import click
from subprocess import call
from colorama import Fore, Back, Style
import importlib
from importlib_metadata import metadata
from tabulate import tabulate
from PyInquirer import prompt, print_json
import platform

class CraftBeerPiCli():
    def __init__(self, config) -> None:
        self.config = config
        pass

    def setup(self):
        print("Setting up CraftBeerPi")
        self.config.create_home_folder_structure()
        self.config.create_config_file()

    def start(self):
        if self.config.check_for_setup() is False:
            return
        print("START")
        cbpi = CraftBeerPi(self.config)
        cbpi.start()

    def setup_one_wire(self):
        print("Setting up 1Wire")
        with open('/boot/config.txt', 'w') as f:
            f.write("dtoverlay=w1-gpio,gpiopin=4,pullup=on")
        print("/boot/config.txt created")

    def list_one_wire(self):
        print("List 1Wire")
        call(["modprobe", "w1-gpio"])
        call(["modprobe", "w1-therm"])
        try:
            for dirname in os.listdir('/sys/bus/w1/devices'):
                if (dirname.startswith("28") or dirname.startswith("10")):
                    print(dirname)
        except Exception as e:
            print(e)

    def plugins_list(self):
        result = []
        print("")
        print(Fore.LIGHTYELLOW_EX,"List of active plugins", Style.RESET_ALL)
        print("")
        discovered_plugins = {
            name: importlib.import_module(name)
            for finder, name, ispkg
            in pkgutil.iter_modules()
            if name.startswith('cbpi') and len(name) > 4
        }
        for key, module in discovered_plugins.items():
            try:
                meta = metadata(key)
                result.append(dict(Name=meta["Name"], Version=meta["Version"], Author=meta["Author"], Homepage=meta["Home-page"], Summary=meta["Summary"]))
                            
            except Exception as e:
                print(e)
        print(Fore.LIGHTGREEN_EX, tabulate(result, headers="keys"), Style.RESET_ALL)


    def plugin_create(self):
        print("Plugin Creation")
        print("")

        questions = [
            {
                'type': 'input',
                'name': 'name',
                'message': 'Plugin Name:',
            }
        ]

        answers = prompt(questions)

        name = "cbpi4_" + answers["name"]
        if os.path.exists(os.path.join(".", name)) is True:
            print("Cant create Plugin. Folder {} already exists ".format(name))
            return

        url = 'https://github.com/Manuel83/craftbeerpi4-plugin-template/archive/main.zip'
        r = requests.get(url)
        with open('temp.zip', 'wb') as f:
            f.write(r.content)

        with ZipFile('temp.zip', 'r') as repo_zip:
            repo_zip.extractall()

        os.rename("./craftbeerpi4-plugin-template-main", os.path.join(".", name))
        os.rename(os.path.join(".", name, "src"), os.path.join(".", name, name))

        import jinja2

        templateLoader = jinja2.FileSystemLoader(searchpath=os.path.join(".", name))
        templateEnv = jinja2.Environment(loader=templateLoader)
        TEMPLATE_FILE = "setup.py"
        template = templateEnv.get_template(TEMPLATE_FILE)
        outputText = template.render(name=name)

        with open(os.path.join(".", name, "setup.py"), "w") as fh:
            fh.write(outputText)

        TEMPLATE_FILE = "MANIFEST.in"
        template = templateEnv.get_template(TEMPLATE_FILE)
        outputText = template.render(name=name)
        with open(os.path.join(".", name, "MANIFEST.in"), "w") as fh:
            fh.write(outputText)

        TEMPLATE_FILE = os.path.join("/", name, "config.yaml")
        operatingsystem = str(platform.system()).lower()
        if operatingsystem.startswith("win"):
            TEMPLATE_FILE=str(TEMPLATE_FILE).replace('\\','/')

        template = templateEnv.get_template(TEMPLATE_FILE)
        outputText = template.render(name=name)

        with open(os.path.join(".", name, name, "config.yaml"), "w") as fh:
            fh.write(outputText)

        print("")
        print("")
        print("Plugin {}{}{} created! ".format(Fore.LIGHTGREEN_EX, name, Style.RESET_ALL) )
        print("")
        print("Developer Documentation: https://openbrewing.gitbook.io/craftbeerpi4_support/readme/development")
        print("")
        print("Happy developing! Cheers")
        print("")
        print("")

    def autostart(self, name):
        '''Enable or disable autostart'''
        if(name == "status"):
            if os.path.exists(os.path.join("/etc/systemd/system","craftbeerpi.service")) is True:
                print("CraftBeerPi Autostart is {}ON{}".format(Fore.LIGHTGREEN_EX,Style.RESET_ALL))
            else:
                print("CraftBeerPi Autostart is {}OFF{}".format(Fore.RED,Style.RESET_ALL))
        elif(name == "on"):
            print("Add craftbeerpi.service to systemd")
            try:
                if os.path.exists(os.path.join("/etc/systemd/system","craftbeerpi.service")) is False:
                    srcfile = self.config.get_file_path("craftbeerpi.service")
                    destfile = os.path.join("/etc/systemd/system")
                    shutil.copy(srcfile, destfile)
                    print("Copied craftbeerpi.service to /etc/systemd/system")
                    os.system('systemctl enable craftbeerpi.service')
                    print('Enabled craftbeerpi service')
                    os.system('systemctl start craftbeerpi.service')
                    print('Started craftbeerpi.service')
                else:
                    print("craftbeerpi.service is already located in /etc/systemd/system")
            except Exception as e:
                print(e)
                return
            return
        elif(name == "off"): 
            print("Remove craftbeerpi.service from systemd")
            try:
                status = os.popen('systemctl list-units --type=service --state=running | grep craftbeerpi.service').read()
                if status.find("craftbeerpi.service") != -1:
                    os.system('systemctl stop craftbeerpi.service')
                    print('Stopped craftbeerpi service')
                    os.system('systemctl disable craftbeerpi.service')
                    print('Removed craftbeerpi.service as service')
                else:
                    print('craftbeerpi.service service is not running')

                if os.path.exists(os.path.join("/etc/systemd/system","craftbeerpi.service")) is True:
                    os.remove(os.path.join("/etc/systemd/system","craftbeerpi.service")) 
                    print("Deleted craftbeerpi.service from /etc/systemd/system")
                else:
                    print("craftbeerpi.service is not located in /etc/systemd/system")
            except Exception as e:
                print(e)
                return
            return


    def chromium(self, name):
        '''Enable or disable autostart'''
        if(name == "status"):
            if os.path.exists(os.path.join("/etc/xdg/autostart/","chromium.desktop")) is True:
                print("CraftBeerPi Chromium Desktop is {}ON{}".format(Fore.LIGHTGREEN_EX,Style.RESET_ALL))
            else:
                print("CraftBeerPi Chromium Desktop is {}OFF{}".format(Fore.RED,Style.RESET_ALL))
        elif(name == "on"):
            print("Add chromium.desktop to /etc/xdg/autostart/")
            try:
                if os.path.exists(os.path.join("/etc/xdg/autostart/","chromium.desktop")) is False:
                    srcfile = self.config.get_file_path("chromium.desktop")
                    destfile = os.path.join("/etc/xdg/autostart/")
                    shutil.copy(srcfile, destfile)
                    print("Copied chromium.desktop to /etc/xdg/autostart/")
                else:
                    print("chromium.desktop is already located in /etc/xdg/autostart/")
            except Exception as e:
                print(e)
                return
            return
        elif(name == "off"): 
            print("Remove chromium.desktop from /etc/xdg/autostart/")
            try:
                if os.path.exists(os.path.join("/etc/xdg/autostart/","chromium.desktop")) is True:
                    os.remove(os.path.join("/etc/xdg/autostart/","chromium.desktop"))
                    print("Deleted chromium.desktop from /etc/xdg/autostart/")
                else:
                    print("chromium.desktop is not located in /etc/xdg/autostart/")
            except Exception as e:
                print(e)
                return
            return


@click.group()
@click.pass_context
@click.option('--config-folder-path', '-c', default="./config", type=click.Path(), help="Specify where the config folder is located. Defaults to './config'.")
def main(context, config_folder_path):
    print("---------------------")
    print("Welcome to CBPi")
    print("---------------------")
    level = logging.INFO
    logging.basicConfig(level=level, format='%(asctime)s - %(levelname)s - %(name)s - %(message)s')
    cbpi_cli = CraftBeerPiCli(ConfigFolder(config_folder_path))
    context.obj = cbpi_cli
    pass

@main.command()
@click.pass_context
def setup(context):
    '''Create Config folder'''
    context.obj.setup()

@main.command()
@click.pass_context
@click.option('--list', is_flag=True, help="List all 1Wire Devices")
@click.option('--setup', is_flag=True, help="Setup 1Wire on Raspberry Pi")
def onewire(context, list, setup):
    '''Setup 1wire on Raspberry Pi'''
    if setup is True:
        context.obj.setup_one_wire()
    if list is True:
        context.obj.list_one_wire()

@main.command()
@click.pass_context
def start(context):
    context.obj.start()

@main.command()
@click.pass_context
def plugins(context):
    '''List active plugins'''
    context.obj.plugins_list()

@main.command()
@click.pass_context
def create(context):
    '''Create New Plugin'''
    context.obj.plugin_create()

@main.command()
@click.pass_context
@click.argument('name')
def autostart(context, name):
    '''(on|off|status) Enable or disable autostart'''
    context.obj.autostart(name)


@main.command()
@click.pass_context
@click.argument('name')
def chromium(context, name):
    '''(on|off|status) Enable or disable Kiosk mode'''
    context.obj.chromium(name)



import logging
import os
import shutil
import psutil
import pathlib
import json
import aiohttp
from voluptuous.schema_builder import message
from cbpi.api.dataclasses import NotificationAction, NotificationType
from cbpi.api.base import CBPiBase
from cbpi.api.config import ConfigType
from cbpi.api import *
import zipfile
import socket

class SystemController:

    def __init__(self, cbpi):
        self.cbpi = cbpi
        self.service = cbpi.actor
        self.logger = logging.getLogger(__name__)

        self.cbpi.app.on_startup.append(self.check_for_update)


    async def check_for_update(self, app):
        pass

    async def restart(self):
        logging.info("RESTART")
        os.system('systemctl reboot') 
        pass

    async def shutdown(self):
        logging.info("SHUTDOWN")
        os.system('systemctl poweroff') 
        pass

    async def backupConfig(self):
        output_filename = "cbpi4_config"
        dir_name = pathlib.Path(self.cbpi.config_folder.get_file_path(''))
        shutil.make_archive(output_filename, 'zip', dir_name)

    async def downloadlog(self, logtime):
        filename = "cbpi4.log"
        fullname = pathlib.Path(os.path.join(".",filename))
        pluginname = "cbpi4_plugins.txt"
        fullpluginname = pathlib.Path(os.path.join(".",pluginname))
        actorname = "cbpi4_actors.txt"
        fullactorname = pathlib.Path(os.path.join(".",actorname))
        sensorname = "cbpi4_sensors.txt"
        fullsensorname = pathlib.Path(os.path.join(".",sensorname))
        kettlename = "cbpi4_kettles.txt"
        fullkettlename = pathlib.Path(os.path.join(".",kettlename))

        output_filename="cbpi4_log.zip"

        if logtime == "b":
            os.system('journalctl -b -u craftbeerpi.service > {}'.format(fullname))
        else:
            os.system('journalctl --since \"{} hours ago\" -u craftbeerpi.service > {}'.format(logtime, fullname))

        os.system('cbpi plugins > {}'.format(fullpluginname))

        try:
            actors = self.cbpi.actor.get_state()
            json.dump(actors['data'],open(fullactorname,'w'),indent=4, sort_keys=True)
            sensors = self.cbpi.sensor.get_state()
            json.dump(sensors['data'],open(fullsensorname,'w'),indent=4, sort_keys=True)
            kettles = self.cbpi.kettle.get_state()
            json.dump(kettles['data'],open(fullkettlename,'w'),indent=4, sort_keys=True)
        except Exception as e:
            logging.info(e)
            self.cbpi.notify("Error", "Creation of files failed: {}".format(e), NotificationType.ERROR)

        try:
            zipObj=zipfile.ZipFile(output_filename , 'w', zipfile.ZIP_DEFLATED)
            zipObj.write(fullname)
            zipObj.write(fullpluginname)
            zipObj.write(fullactorname)
            zipObj.write(fullsensorname)
            zipObj.write(fullkettlename)
            zipObj.close()
        except Exception as e:
            logging.info(e)
            self.cbpi.notify("Error", "Zip creation failed: {}".format(e), NotificationType.ERROR)

        try:
            os.remove(fullname)
            os.remove(fullpluginname)
            os.remove(fullactorname)
            os.remove(fullsensorname)
            os.remove(fullkettlename)
        except Exception as e:
            logging.info(e)
            self.cbpi.notify("Error", "Removal of original files failed: {}".format(e), NotificationType.ERROR)


    def allowed_file(self, filename, extension):
        return '.' in filename and filename.rsplit('.', 1)[1] in set([extension])

    def recursive_chown(self, path, owner, group):
        for dirpath, dirnames, filenames in os.walk(path):
            shutil.chown(dirpath, owner, group)
            for filename in filenames:
                shutil.chown(os.path.join(dirpath, filename), owner, group)

    async def restoreConfig(self, data):
        fileData = data['File']
        filename = fileData.filename
        backup_file = fileData.file
        content_type = fileData.content_type
        required_content=['dashboard/', 'recipes/', 'upload/', 'config.json', 'config.yaml']

        if content_type == 'application/x-zip-compressed':
            try:
                content = backup_file.read()
                if backup_file and self.allowed_file(filename, 'zip'):
                    self.path = os.path.join(".", "restored_config.zip")

                    f=open(self.path, "wb")
                    f.write(content)
                    f.close()
                    zip=zipfile.ZipFile(self.path)
                    zip_content_list = zip.namelist()
                    zip_content = True
                    for content in required_content:
                        try:
                            check = zip_content_list.index(content)
                        except:
                            zip_content = False
                    if zip_content == True:
                        self.cbpi.notify("Success", "Config backup has been uploaded", NotificationType.SUCCESS)
                        self.cbpi.notify("Action Required!", "Please restart the server", NotificationType.WARNING)
                    else:
                        self.cbpi.notify("Error", "Wrong content type. Upload failed", NotificationType.ERROR)
                        os.remove(self.path)
            except:
                self.cbpi.notify("Error", "Config backup upload failed", NotificationType.ERROR)
                pass
        else:
            self.cbpi.notify("Error", "Wrong content type. Upload failed", NotificationType.ERROR)

    async def uploadSVG(self, data):
        fileData = data['File']
        filename = fileData.filename
        svg_file = fileData.file
        content_type = fileData.content_type

        logging.info(content_type)

        if content_type == 'image/svg+xml':
            try:
                content = svg_file.read().decode('utf-8','replace')
                if svg_file and self.allowed_file(filename, 'svg'):
                    self.path = os.path.join(self.cbpi.config_folder.get_file_path("dashboard"),"widgets", filename)
                    logging.info(self.path)

                    f=open(self.path, "w")
                    f.write(content)
                    f.close()
                    self.cbpi.notify("Success", "SVG file ({}) has been uploaded.".format(filename), NotificationType.SUCCESS)
            except:
                self.cbpi.notify("Error", "SVG upload failed", NotificationType.ERROR)
                pass
        else:
            self.cbpi.notify("Error", "Wrong content type. Upload failed", NotificationType.ERROR)

    async def systeminfo(self):
        logging.info("SYSTEMINFO")
        system = "" 
        temp = 0
        cpuload = 0
        cpucount = 0
        cpufreq = 0
        totalmem = 0
        availmem = 0
        mempercent = 0
        eth0IP = "N/A"
        wlan0IP = "N/A"
        eth0speed = "N/A"
        wlan0speed = "N/A"       

        TEMP_UNIT=self.cbpi.config.get("TEMP_UNIT", "C")
        FAHRENHEIT = False if TEMP_UNIT == "C" else True

        af_map = { socket.AF_INET: 'IPv4',
                   socket.AF_INET6: 'IPv6',
                   }

        try:
            if psutil.LINUX == True:
                system = "Linux"
            elif psutil.WINDOWS == True:
                system = "Windows"
            elif psutil.MACOS == True:
                system = "MacOS"
            cpuload = round(psutil.cpu_percent(interval=None),1)
            cpucount = psutil.cpu_count(logical=False)
            cpufreq = psutil.cpu_freq()
            mem = psutil.virtual_memory()
            availmem = round((int(mem.available) / (1024*1024)),1)
            mempercent = round(float(mem.percent),1)
            totalmem = round((int(mem.total) / (1024*1024)),1)
            if system == "Linux":
                try:
                    temps = psutil.sensors_temperatures(fahrenheit=FAHRENHEIT)
                    for name, entries in temps.items():
                        for entry in entries:
                            if name == "cpu_thermal":
                                temp = round(float(entry.current),1)
                except:
                    pass
            else:
                temp = "N/A"
            if system == "Linux":
                try:
                    ethernet = psutil.net_if_addrs()
                    for nic, addrs in ethernet.items():
                        if nic == "eth0":
                            for addr in addrs:
                                if str(addr.family) == "AddressFamily.AF_INET": 
                                    if addr.address:
                                        eth0IP = addr.address
                        if nic == "wlan0":
                            for addr in addrs:
                                if str(addr.family) == "AddressFamily.AF_INET": 
                                    if addr.address:
                                        wlan0IP = addr.address
                    info = psutil.net_if_stats()
                    try:
                        for nic in info:
                            if nic == 'eth0':
                                if info[nic].isup == True:
                                    if info[nic].speed:
                                        eth0speed = info[nic].speed
                                else:
                                    eth0speed = "down"
                            if nic == 'wlan0':
                                if info[nic].isup == True: 
                                    ratestring = os.popen('iwlist wlan0 rate | grep Rate').read()
                                    start = ratestring.find("=") + 1
                                    end = ratestring.find(" Mb/s")
                                    wlan0speed = ratestring[start:end]
                                else:
                                    wlan0speed = "down"
                    except Exception as e:
                        logging.info(e)
                except:
                    pass

            if system == "Windows":
                try:
                    ethernet = psutil.net_if_addrs()               
                    for nic, addrs in ethernet.items():
                        if nic == "Ethernet":
                            for addr in addrs:
                                if str(addr.family) == "AddressFamily.AF_INET": 
                                    if addr.address:
                                        eth0IP = addr.address
                        if nic == "WLAN":
                            for addr in addrs:
                                if str(addr.family) == "AddressFamily.AF_INET": 
                                    if addr.address:
                                        wlan0IP = addr.address
                    info = psutil.net_if_stats()
                    try:
                        for nic in info:
                            if nic == 'Ethernet':
                                if info[nic].isup == True:
                                    if info[nic].speed:
                                        eth0speed = info[nic].speed
                                else:
                                    eth0speed = "down"
                            if nic == 'WLAN':
                                if info[nic].isup == True:
                                    if info[nic].speed:
                                        wlan0speed = info[nic].speed
                                else:
                                    wlan0speed = "down"                    
                    except Exception as e:
                        logging.info(e)
                except:
                    pass

        except:
            pass

        systeminfo =    {'system': system,
                         'cpuload': cpuload,
                         'cpucount': cpucount,
                         'cpufreq': cpufreq.current,
                         'totalmem': totalmem,
                         'availmem': availmem,
                         'mempercent': mempercent,
                         'temp': temp,
                         'temp_unit': TEMP_UNIT,
                         'eth0': eth0IP,
                         'wlan0': wlan0IP,
                         'eth0speed': eth0speed,
                         'wlan0speed': wlan0speed}
        return systeminfo





from functools import wraps

from aiohttp.web import View

from . import create_scheduler

__all__ = ('setup', 'spawn', 'get_scheduler', 'get_scheduler_from_app', 'atomic')


def get_scheduler(request):
    scheduler = get_scheduler_from_request(request)
    if scheduler is None:
        raise RuntimeError(
            "Call aiojobs.aiohttp.setup() on application initialization")
    return scheduler


def get_scheduler_from_app(app):
    return app.get('AIOJOBS_SCHEDULER')


def get_scheduler_from_request(request):
    return request.config_dict.get('AIOJOBS_SCHEDULER')


async def spawn(request, coro, name="Manuel"):
    return await get_scheduler(request).spawn(coro)


def atomic(coro):
    @wraps(coro)
    async def wrapper(request):
        if isinstance(request, View):
            # Class Based View decorated.
            request = request.request

        job = await spawn(request, coro(request))
        return await job.wait()
    return wrapper



async def setup(app, cbpi, **kwargs):

    app['AIOJOBS_SCHEDULER'] = await create_scheduler(cbpi, **kwargs)

    async def on_cleanup(app):
        await app['AIOJOBS_SCHEDULER'].close()

    app.on_cleanup.append(on_cleanup)





from socket import timeout
from typing import KeysView

from voluptuous.schema_builder import message
from cbpi.api.dataclasses import NotificationAction, NotificationType
import logging
from unittest.mock import MagicMock, patch
from datetime import datetime
from cbpi.api import *

logger = logging.getLogger(__name__)

@parameters([])
class DummyActor(CBPiActor):
    

    def __init__(self, cbpi, id, props):
        super().__init__(cbpi, id, props)
    
    @action("SAY HELLO", {})
    async def helloWorld(self, **kwargs):
        self.cbpi.notify("HELLO", "WOOHO", NotificationType.ERROR)

    async def start(self):
        await super().start()

    async def on(self, power=0):
        logger.info("ACTOR %s ON " %  self.id)
        self.state = True

    async def off(self):
        logger.info("ACTOR %s OFF " % self.id)
        
        self.state = False

    def get_state(self):
        
        return self.state
    
    async def run(self):
        pass

def setup(cbpi):

    '''
    This method is called by the server during startup 
    Here you need to register your plugins at the server
    
    :param cbpi: the cbpi core 
    :return: 
    '''

    cbpi.plugin.register("DummyActor", DummyActor)



import time

from aiohttp.test_utils import unittest_run_loop
from tests.cbpi_config_fixture import CraftBeerPiTestCase

class ConfigTestCase(CraftBeerPiTestCase):

    @unittest_run_loop
    async def test_get(self):

        assert self.cbpi.config.get("steps_boil_temp", 1) == "99"

    @unittest_run_loop
    async def test_set_get(self):
        value = 35

        await self.cbpi.config.set("steps_cooldown_temp", value)
        assert self.cbpi.config.get("steps_cooldown_temp", 1) == value

    @unittest_run_loop
    async def test_http_set(self):
        value = "Some New Brewery Name"
        key = "BREWERY_NAME"

        resp = await self.client.request("PUT", "/config/%s/" % key, json={'value': value})
        assert resp.status == 204

        assert self.cbpi.config.get(key, -1) == value

    @unittest_run_loop
    async def test_http_get(self):
        resp = await self.client.request("GET", "/config/")
        assert resp.status == 200

    @unittest_run_loop
    async def test_get_default(self):
        value = self.cbpi.config.get("HELLO_WORLD", "DefaultValue")
        assert value == "DefaultValue"


# -*- coding: utf-8 -*-
import asyncio
import random
import re
import random
from aiohttp import web
from cbpi.api import *
import os, re, threading, time
from subprocess import call
import random

def getSensors():
    try:
        arr = []
        for dirname in os.listdir('/sys/bus/w1/devices'):
            if (dirname.startswith("28") or dirname.startswith("10")):
                arr.append(dirname)
        return arr
    except:
        return []


class ReadThread (threading.Thread):

    value = 0
    def __init__(self, sensor_name):
        threading.Thread.__init__(self)
        self.value = 0
        self.sensor_name = sensor_name
        self.runnig = True

    def shutdown(self):
        pass

    def stop(self):
        self.runnig = False

    def run(self):

        while self.runnig:
            try:                
                if self.sensor_name is None:
                    return
                with open('/sys/bus/w1/devices/%s/w1_slave' % self.sensor_name, 'r') as content_file:
                    content = content_file.read()
                    if (content.split('\n')[0].split(' ')[11] == "YES"):
                        temp = float(content.split("=")[-1]) / 1000  # temp in Celcius
                        self.value = temp
            except:
                pass
            
            time.sleep(1)

@parameters([Property.Select(label="Sensor", options=getSensors()), 
             Property.Number(label="offset",configurable = True, default_value = 0, description="Sensor Offset (Default is 0)"),
             Property.Select(label="Interval", options=[1,5,10,30,60], description="Interval in Seconds")])
class OneWire(CBPiSensor):
    
    def __init__(self, cbpi, id, props):
        super(OneWire, self).__init__(cbpi, id, props)
        self.value = 200

    async def start(self):
        await super().start()
        self.name = self.props.get("Sensor")
        self.interval = self.props.get("Interval", 60)
        self.offset = float(self.props.get("offset",0))

        self.t = ReadThread(self.name)
        self.t.daemon = True
        def shudown():
            shudown.cb.shutdown()
        shudown.cb = self.t
        self.t.start()
    
    async def stop(self):
        try:
            self.t.stop()
            self.running = False
        except:
            pass

    async def run(self):
        while self.running == True:
            self.TEMP_UNIT=self.get_config_value("TEMP_UNIT", "C")
            if self.TEMP_UNIT == "C": # Report temp in C if nothing else is selected in settings
                self.value = round((self.t.value + self.offset),2)
            else: # Report temp in F if unit selected in settings
                self.value = round((9.0 / 5.0 * self.t.value + 32 + self.offset), 2)
            
            self.log_data(self.value)
            self.push_update(self.value)
            await asyncio.sleep(self.interval)
    
    def get_state(self):
        return dict(value=self.value)


def setup(cbpi):
    cbpi.plugin.register("OneWire", OneWire)
    try:
        # Global Init
        call(["modprobe", "w1-gpio"])
        call(["modprobe", "w1-therm"])
    except Exception as e:
        pass




import asyncio
import sys
try:
    from asyncio import set_event_loop_policy, WindowsSelectorEventLoopPolicy
except ImportError:
    pass
import json
from voluptuous.schema_builder import message
from cbpi.api.dataclasses import NotificationType
from cbpi.controller.notification_controller import NotificationController
import logging
from os import urandom
import os
from cbpi import __version__, __codename__
from aiohttp import web
from aiohttp_auth import auth
from aiohttp_session import session_middleware
from aiohttp_session.cookie_storage import EncryptedCookieStorage
from aiohttp_swagger import setup_swagger
from cbpi.api.exceptions import CBPiException
from voluptuous import MultipleInvalid

from cbpi.controller.dashboard_controller import DashboardController
from cbpi.controller.job_controller import JobController
from cbpi.controller.actor_controller import ActorController
from cbpi.controller.config_controller import ConfigController
from cbpi.controller.kettle_controller import KettleController
from cbpi.controller.plugin_controller import PluginController
from cbpi.controller.sensor_controller import SensorController
from cbpi.controller.step_controller import StepController
from cbpi.controller.recipe_controller import RecipeController
from cbpi.controller.fermenter_recipe_controller import FermenterRecipeController
from cbpi.controller.upload_controller import UploadController
from cbpi.controller.fermentation_controller import FermentationController

from cbpi.controller.system_controller import SystemController
from cbpi.controller.satellite_controller import SatelliteController

from cbpi.controller.log_file_controller import LogController

from cbpi.eventbus import CBPiEventBus
from cbpi.http_endpoints.http_login import Login
from cbpi.utils import *
from cbpi.websocket import CBPiWebSocket
from cbpi.http_endpoints.http_actor import ActorHttpEndpoints

from cbpi.http_endpoints.http_config import ConfigHttpEndpoints
from cbpi.http_endpoints.http_dashboard import DashBoardHttpEndpoints
from cbpi.http_endpoints.http_kettle import KettleHttpEndpoints
from cbpi.http_endpoints.http_sensor import SensorHttpEndpoints
from cbpi.http_endpoints.http_step import StepHttpEndpoints
from cbpi.http_endpoints.http_recipe import RecipeHttpEndpoints
from cbpi.http_endpoints.http_fermenterrecipe import FermenterRecipeHttpEndpoints
from cbpi.http_endpoints.http_plugin import PluginHttpEndpoints
from cbpi.http_endpoints.http_system import SystemHttpEndpoints
from cbpi.http_endpoints.http_log import LogHttpEndpoints
from cbpi.http_endpoints.http_notification import NotificationHttpEndpoints
from cbpi.http_endpoints.http_upload import UploadHttpEndpoints
from cbpi.http_endpoints.http_fermentation import FermentationHttpEndpoints

import shortuuid
logger = logging.getLogger(__name__)

@web.middleware
async def error_middleware(request, handler):
    try:
        response = await handler(request)
        if response.status != 404:
            return response
        message = response.message
    except web.HTTPException as ex:
        if ex.status != 404:
            raise
        message = ex.reason
    except CBPiException as ex:
        message = str(ex)
        return web.json_response(status=500, data={'error': message})
    except MultipleInvalid as ex:
        return web.json_response(status=500, data={'error': str(ex)})
    except Exception as ex:
        return web.json_response(status=500, data={'error': str(ex)})

    return web.json_response(status=500, data={'error': message})


class CraftBeerPi:

    def __init__(self, configFolder):

        operationsystem= sys.platform
        if operationsystem.startswith('win'):
            set_event_loop_policy(WindowsSelectorEventLoopPolicy())

        self.path = os.sep.join(os.path.abspath(__file__).split(os.sep)[:-1])  # The path to the package dir
        
        self.version = __version__
        self.codename = __codename__

        self.config_folder = configFolder
        self.static_config = load_config(configFolder.get_file_path("config.yaml"))
        
        logger.info("Init CraftBeerPI")

        policy = auth.SessionTktAuthentication(urandom(32), 60, include_ip=True)
        middlewares = [web.normalize_path_middleware(), session_middleware(EncryptedCookieStorage(urandom(32))),
                       auth.auth_middleware(policy), error_middleware]
        # max upload size increased to 5 Mb. Default is 1 Mb -> config and svg upload
        self.app = web.Application(middlewares=middlewares, client_max_size=5*1024*1024)
        self.app["cbpi"] = self

        self._setup_shutdownhook()
        self.initializer = []

        self.bus = CBPiEventBus(self.app.loop, self)
        self.job = JobController(self)
        self.config = ConfigController(self)
        self.ws = CBPiWebSocket(self)
        self.actor = ActorController(self)
        self.sensor = SensorController(self)
        self.plugin = PluginController(self)
        self.log = LogController(self)
        self.system = SystemController(self)
        self.kettle = KettleController(self)
        self.fermenter : FermentationController = FermentationController(self)
        self.step : StepController = StepController(self)
        self.recipe : RecipeController = RecipeController(self)
        self.fermenterrecipe : FermenterRecipeController = FermenterRecipeController(self)
        self.upload : UploadController = UploadController(self)
        self.notification : NotificationController = NotificationController(self)
        self.satellite = None
        if str(self.static_config.get("mqtt", False)).lower() == "true":
            self.satellite: SatelliteController = SatelliteController(self)
        self.dashboard = DashboardController(self)

        self.http_step = StepHttpEndpoints(self)
        self.http_recipe = RecipeHttpEndpoints(self)
        self.http_fermenterrecipe = FermenterRecipeHttpEndpoints(self)
        self.http_sensor = SensorHttpEndpoints(self)
        self.http_config = ConfigHttpEndpoints(self)
        self.http_actor = ActorHttpEndpoints(self)
        self.http_kettle = KettleHttpEndpoints(self)
        self.http_dashboard = DashBoardHttpEndpoints(self)
        self.http_plugin = PluginHttpEndpoints(self)
        self.http_system = SystemHttpEndpoints(self)
        self.http_log = LogHttpEndpoints(self)
        self.http_notification = NotificationHttpEndpoints(self)
        self.http_upload = UploadHttpEndpoints(self)
        self.http_fermenter = FermentationHttpEndpoints(self)

        self.login = Login(self)

    def _setup_shutdownhook(self):
        self.shutdown = False

        async def on_cleanup(app):
            self.shutdown = True

        self.app.on_cleanup.append(on_cleanup)

    def register_on_startup(self, obj):

        for method in [getattr(obj, f) for f in dir(obj) if
                       callable(getattr(obj, f)) and hasattr(getattr(obj, f), "on_startup")]:
            name = method.__getattribute__("name")
            order = method.__getattribute__("order")
            self.initializer.append(dict(name=name, method=method, order=order))

    def register(self, obj, url_prefix=None, static=None):

        '''
        This method parses the provided object
        
        :param obj: the object wich will be parsed for registration 
        :param url_prefix: that prefix for HTTP Endpoints
        :return: None
        '''
        self.register_http_endpoints(obj, url_prefix, static)
        self.bus.register_object(obj)
        # self.ws.register_object(obj)
        self.job.register_background_task(obj)
        self.register_on_startup(obj)

    def register_http_endpoints(self, obj, url_prefix=None, static=None):

        if url_prefix is None:
            logger.debug(
                "URL Prefix is None for %s. No endpoints will be registered. Please set / explicit if you want to add it to the root path" % obj)
            return
        routes = []
        for method in [getattr(obj, f) for f in dir(obj) if
                       callable(getattr(obj, f)) and hasattr(getattr(obj, f), "route")]:
            http_method = method.__getattribute__("method")
            path = method.__getattribute__("path")
            class_name = method.__self__.__class__.__name__
            logger.debug(
                "Register Endpoint : %s.%s %s %s%s " % (class_name, method.__name__, http_method, url_prefix, path))

            def add_post():
                routes.append(web.post(method.__getattribute__("path"), method))

            def add_get():
                routes.append(web.get(method.__getattribute__("path"), method, allow_head=False))

            def add_delete():
                routes.append(web.delete(path, method))

            def add_put():
                routes.append(web.put(path, method))

            switcher = {
                "POST": add_post,
                "GET": add_get,
                "DELETE": add_delete,
                "PUT": add_put
            }
            switcher[http_method]()

        if url_prefix != "/":
            logger.debug("URL Prefix: %s " % (url_prefix,))
            sub = web.Application()
            sub.add_routes(routes)
            if static is not None:
                sub.add_routes([web.static('/static', static, show_index=True)])
            self.app.add_subapp(url_prefix, sub)
        else:

            self.app.add_routes(routes)

    def _swagger_setup(self):
        '''
        Internal method to expose REST API documentation by swagger
        
        :return: 
        '''
        long_description = """
        This is the api for CraftBeerPi
        """
        setup_swagger(self.app,
                      description=long_description,
                      title="CraftBeerPi",
                      api_version=self.version,
                      contact="info@craftbeerpi.com")


    def notify(self, title: str, message: str, type: NotificationType = NotificationType.INFO, action=[]) -> None:
        self.notification.notify(title, message, type, action)
        
    def push_update(self, topic, data, retain=False) -> None:

        if self.satellite is not None:
            asyncio.create_task(self.satellite.publish(topic=topic, message=json.dumps(data), retain=retain))

    async def call_initializer(self, app):
        self.initializer = sorted(self.initializer, key=lambda k: k['order'])
        for i in self.initializer:
            logger.info("CALL INITIALIZER %s - %s " % (i["name"], i["method"].__name__))
            await i["method"]()

    def _print_logo(self):
        from pyfiglet import Figlet
        f = Figlet(font='big')
        logger.info("\n%s" % f.renderText("CraftBeerPi %s " % self.version))
        logger.info("www.CraftBeerPi.com")
        logger.info("(c) 2021/2022 Manuel Fritsch / Alexander Vollkopf")

    def _setup_http_index(self):
        async def http_index(request):
            url = self.config.static.get("index_url")

            if url is not None:

                raise web.HTTPFound(url)
            else:
                return web.Response(text="Hello from CraftbeerPi!")

        self.app.add_routes([web.get('/', http_index),
                             web.static('/static', os.path.join(os.path.dirname(__file__), "static"), show_index=True)])

    async def init_serivces(self):

        self._print_logo()

        await self.job.init()
        
        await self.config.init()
        if self.satellite is not None:
            await self.satellite.init()
        self._setup_http_index()
        self.plugin.load_plugins()
        self.plugin.load_plugins_from_evn()
        await self.fermenter.init()
        await self.sensor.init()
        await self.step.init()
        
        await self.actor.init()
        await self.kettle.init()
        await self.call_initializer(self.app)
        await self.dashboard.init()
    

        self._swagger_setup()

        level = logging.INFO
        logger = logging.getLogger()
        logger.setLevel(level)
        for handler in logger.handlers:
            handler.setLevel(level)

        return self.app

    def start(self):
        web.run_app(self.init_serivces(), port=self.static_config.get("port", 2202))



# -*- coding: utf-8 -*-
import asyncio
import random
import logging
from cbpi.api import *
from cbpi.api.base import CBPiBase
from cbpi.api.dataclasses import Kettle, Props, Fermenter

@parameters([])
class CustomSensor(CBPiSensor):

    def __init__(self, cbpi, id, props):
        super(CustomSensor, self).__init__(cbpi, id, props)
        self.value = 0

    async def run(self):
        while self.running:
            self.value = random.randint(10, 100)
            self.log_data(self.value)

            self.push_update(self.value)
            await asyncio.sleep(1)

    def get_state(self):
        return dict(value=self.value)

@parameters([Property.Number(label="Pressure", configurable=True, description="Start Pressure"),
             Property.Number(label="PressureIncrease", configurable=True, description="Pressure increase per hour"),
             Property.Number(label="PressureDecrease", configurable=True, description="Pressure decrease per second on openm valve"),
             Property.Fermenter(label="Fermenter",description="Fermenter")])
class DummyPressure(CBPiSensor):

    def __init__(self, cbpi, id, props):
        super(DummyPressure, self).__init__(cbpi, id, props)
        self.value = float(self.props.get("Pressure",0))
        fermenter=self.props.get("Fermenter",None)
        self.fermenter=self.get_fermenter(fermenter)
        self.valve=self.fermenter.valve

    async def run(self):
        self.uprate=float(self.props.get("PressureIncrease",0))/3600
        self.decrease=float(self.props.get("PressureDecrease",0))
        logging.info(self.uprate)
        logging.info(self.decrease)

        while self.running:
            valve_state=self.get_actor_state(self.valve)
            fermenter_instance=self.fermenter.instance
            if fermenter_instance:
                fermenter_state=fermenter_instance.state
            else:
                fermenter_state = False
            if valve_state == False and fermenter_state:
                self.value = self.value + self.uprate
            elif valve_state and fermenter_state:
                self.value=self.value-self.decrease

            self.log_data(self.value)

            self.push_update(round(self.value,2))
            await asyncio.sleep(1)

    def get_state(self):
        return dict(value=self.value)

def setup(cbpi):
    '''
    This method is called by the server during startup
    Here you need to register your plugins at the server

    :param cbpi: the cbpi core
    :return:
    '''
    cbpi.plugin.register("CustomSensor", CustomSensor)
    cbpi.plugin.register("DummyPressure", DummyPressure)



from cbpi.api.config import ConfigType
from enum import Enum
from typing import Any
from cbpi.api.step import StepState
from dataclasses import dataclass, field
from typing import List

class Props:

    def __init__(self, data={}):
        super(Props, self).__setattr__('__data__', {})
        for key, value in data.items():
            self.__setattr__(key, value)
    def __getattr__(self, name):
        return self.__data__.get(name)
    
    def __setattr__(self, name, value):
        self.__data__[name] = value

    def __str__(self):
        return self.__data__.__str__()


    def __getitem__(self, key):
        return self.__data__[key]

    def __setitem__(self, key, value):
        self.__data__[key] = value

    def __contains__(self, key):
        return key in self.__data__

    def get(self, key, d=None):
        if key in self.__data__ and self.__data__[key] != "": 
            return self.__data__[key]
        else:
            return d
    

    def to_dict(self):
        def parse_object(value):
            if isinstance(value, Props):
                return value.to_dict()
            elif isinstance(value, list):
                return list(map(parse_object, value))
            else:
                return value

        return dict((key, parse_object(value)) for (key, value) in self.__data__.items())


@dataclass
class Actor:
    id: str = None
    name: str = None
    props: Props = Props()
    state: bool = False
    power: int = 100
    type: str = None
    instance: str = None

    def __str__(self):
        return "name={} props={}, state={}, type={}, power={}".format(self.name, self.props, self.state, self.type, self.power)
    def to_dict(self):
        return dict(id=self.id, name=self.name, type=self.type, props=self.props.to_dict(), state=self.instance.get_state(), power=self.power)


@dataclass
class Sensor:
    id: str = None
    name: str = None
    props: Props = Props()
    state: bool = False
    type: str = None
    instance: str = None

    def __str__(self):
        return "name={} props={}, state={}".format(self.name, self.props, self.state)
    def to_dict(self):
        return dict(id=self.id, name=self.name, type=self.type, props=self.props.to_dict(), state=self.state)

@dataclass
class Kettle:
    id: str = None
    name: str = None
    props: Props = Props()
    instance: str = None
    agitator: Actor = None
    heater: Actor = None
    sensor: Sensor = None
    type: str = None
    target_temp: int = 0

    def __str__(self):
        return "name={} props={} temp={}".format(self.name, self.props, self.target_temp)
    def to_dict(self):

        if self.instance is not None:
            
            state = self.instance.state

        else:
            state = False
        return dict(id=self.id, name=self.name, state=state,  target_temp=self.target_temp, heater=self.heater, agitator=self.agitator, sensor=self.sensor, type=self.type, props=self.props.to_dict())



@dataclass
class Step:
    id: str = None
    name: str = None
    props: Props = Props()
    type: str = None
    status: StepState = StepState.INITIAL
    instance: str = None

    def __str__(self):
        return "name={} props={}, type={}, instance={}".format(self.name, self.props, self.type, self.instance)
    def to_dict(self):

        msg = self.instance.summary if self.instance is not None else ""
        return dict(id=self.id, name=self.name, state_text=msg, type=self.type, status=self.status.value, props=self.props.to_dict())

@dataclass
class Fermenter:
    id: str = None
    name: str = None
    sensor: Sensor = None
    pressure_sensor : Sensor = None
    heater: Actor = None
    cooler: Actor = None
    valve: Actor = None
    brewname: str = None
    description : str = None
    props: Props = Props()
    target_temp: float = 0 
    target_pressure: float = 0
    type: str = None
    steps: List[Step]= field(default_factory=list)
    instance: str = None


    def __str__(self):
        return "name={} props={} temp={}".format(self.name, self.props, self.target_temp, self.target_pressure)

    def to_dict(self):

        if self.instance is not None:
            
            state = self.instance.state

        else:
            state = False

        steps = list(map(lambda item: item.to_dict(), self.steps))
        return dict(id=self.id, name=self.name, state=state, sensor=self.sensor, pressure_sensor=self.pressure_sensor, heater=self.heater, cooler=self.cooler, valve=self.valve, brewname=self.brewname, description=self.description, props=self.props.to_dict() if self.props is not None else None, target_temp=self.target_temp, target_pressure=self.target_pressure, type=self.type, steps=steps)


@dataclass
class FermenterStep:
    id: str = None
    name: str = None
    fermenter: Fermenter = None
    props: Props = Props()
    type: str = None
    status: StepState = StepState.INITIAL
    endtime: int = 0 # endtime if step is active and timer is running
    instance: str = None
    step: dict = None 

    def __str__(self):
        return "name={} props={}, type={}, instance={}".format(self.name, self.props, self.type, self.instance)
    def to_dict(self):
        msg = self.instance.summary if self.instance is not None else ""
        return dict(id=self.id, name=self.name, state_text=msg, type=self.type, status=self.status.value, endtime=self.endtime, props=self.props.to_dict())



class ConfigType(Enum):
    STRING="string"
    ACTOR="actor"
    SENSOR="sensor"
    KETTLE="kettle"
    NUMBER="number"
    SELECT="select"
    STEP="step"
    FERMENTER="fermenter"

@dataclass  
class Config:
    
    name: str = None
    value: Any = None
    description: str = None
    type: ConfigType = ConfigType.STRING
    options: Any = None

    def __str__(self):
        return "....name={} value={}".format(self.name, self.value)
    def to_dict(self):  
        return dict(name=self.name, value=self.value, type=self.type.value, description=self.description, options=self.options)

@dataclass  
class NotificationAction:
    label: str 
    method: Any = None
    id: str = None

    def to_dict(self):  
        return dict(id=self.id, label=self.label)

class NotificationType(Enum):
    INFO="info"
    WARNING="warning"
    ERROR="error"
    SUCCESS="success"

    def __str__(self):
        return self.value



import unittest
from unittest import mock
from unittest.mock import MagicMock, patch

try:
    import RPi.GPIO as GPIO
except Exception:
    print("Error importing RPi.GPIO!")
    MockRPi = MagicMock()
    modules = {
        "RPi": MockRPi,
        "RPi.GPIO": MockRPi.GPIO
    }
    patcher = patch.dict("sys.modules", modules)
    patcher.start()
    import RPi.GPIO as GPIO

class TestSwitch(unittest.TestCase):

    GPIO_NUM = 22

    @patch("RPi.GPIO.setup")
    def test_switch_inits_gpio(self, patched_setup):
        GPIO.setmode(GPIO.BCM)
        GPIO.setup(self.GPIO_NUM, GPIO.OUT)
        patched_setup.assert_called_once_with(self.GPIO_NUM, GPIO.OUT)

    @patch("RPi.GPIO.output")
    def test_switch_without_scheduler_starts_disabled(self, patched_output):
        GPIO.output(self.GPIO_NUM, GPIO.LOW)
        patched_output.assert_called_once_with(self.GPIO_NUM, GPIO.LOW)


# -*- coding: utf-8 -*-
import asyncio

from cbpi.api import parameters, Property, CBPiSensor
from cbpi.api import *
import logging
import json

@parameters([Property.Text(label="Topic", configurable=True, description="MQTT Topic"),
             Property.Text(label="PayloadDictionary", configurable=True, default_value="",
                           description="Where to find msg in payload, leave blank for raw payload")])
class MQTTSensor(CBPiSensor):

    def __init__(self, cbpi, id, props):
        super(MQTTSensor, self).__init__(cbpi, id, props)
        self.Topic = self.props.get("Topic", None)
        self.payload_text = self.props.get("PayloadDictionary", None)
        if self.payload_text != None:
            self.payload_text = self.payload_text.split('.')
        self.mqtt_task = self.cbpi.satellite.subcribe(self.Topic, self.on_message)
        self.value: float = 999

    async def on_message(self, message):
        val = json.loads(message)
        try:
            if self.payload_text is not None:
                for key in self.payload_text:
                    val = val.get(key, None)

            if isinstance(val, (int, float, str)):
                self.value = float(val)
                self.log_data(self.value)
                self.push_update(self.value)
        except Exception as e:
            logging.info("MQTT Sensor Error {}".format(e))

    async def run(self):
        while self.running:
            await asyncio.sleep(1)

    def get_state(self):
        return dict(value=self.value)

    async def on_stop(self):
        if self.mqtt_task.done() is False:
            self.mqtt_task.cancel()
            try:
                await self.mqtt_task
            except asyncio.CancelledError:
                pass


def setup(cbpi):
    '''
    This method is called by the server during startup
    Here you need to register your plugins at the server

    :param cbpi: the cbpi core
    :return:
    '''
    if str(cbpi.static_config.get("mqtt", False)).lower() == "true":
        cbpi.plugin.register("MQTTSensor", MQTTSensor)



from cbpi.api import parameters, Property
from . import MQTTActor

@parameters([
    Property.Text(label="Topic", configurable=True, description = "MQTT Topic"),
    Property.Text(label="Payload", configurable=True, description = "Payload that is sent as MQTT message. Available placeholders are {switch_onoff}: [on|off], {switch_10}: [1|0], {power}: [0-100].")
])
class GenericMqttActor(MQTTActor):
    def __init__(self, cbpi, id, props):
        MQTTActor.__init__(self, cbpi, id, props)
        self.payload = ""

    async def on_start(self):
         await MQTTActor.on_start(self)
         self.payload = self.props.get("Payload", "{{\"state\": \"{switch_onoff}\", \"power\": {power}}}")

    def normalize_power_value(self, power):
        if power is not None:
            if power != self.power:
                power = min(100, power)
                power = max(0, power)
                self.power = round(power)
    
    async def publish_mqtt_message(self, topic, payload):
        self.logger.info("Publish '{payload}' to '{topic}'".format(payload = payload, topic = self.topic))
        await self.cbpi.satellite.publish(self.topic, payload, True)

    async def on(self, power=None):
        self.normalize_power_value(power)
        formatted_payload = self.payload.format(switch_onoff = "on", switch_10 = 1, power = self.power)
        await self.publish_mqtt_message(self.topic, formatted_payload)
        self.state = True

    async def off(self):
        formatted_payload = self.payload.format(switch_onoff = "off", switch_10 = 0, power = self.power)
        await self.publish_mqtt_message(self.topic, formatted_payload)
        self.state = False


import asyncio

from cbpi.api import parameters, Property, action
from cbpi.api.step import StepResult, CBPiStep
from cbpi.api.timer import Timer
from datetime import datetime
import time
from voluptuous.schema_builder import message
from cbpi.api.dataclasses import NotificationAction, NotificationType
from cbpi.api.dataclasses import Kettle, Props
from cbpi.api import *
import logging
from socket import timeout
from typing import KeysView
from cbpi.api.config import ConfigType
from cbpi.api.base import CBPiBase
import numpy as np
import warnings



@parameters([Property.Text(label="Notification",configurable = True, description = "Text for notification"),
             Property.Select(label="AutoNext",options=["Yes","No"], description="Automatically move to next step (Yes) or pause after Notification (No)")])
class NotificationStep(CBPiStep):

    async def NextStep(self, **kwargs):
        await self.next()

    async def on_timer_done(self,timer):
        self.summary = self.props.get("Notification","")

        if self.AutoNext == True:
            self.cbpi.notify(self.name, self.props.get("Notification",""), NotificationType.INFO)
            await self.next()
        else:
            self.cbpi.notify(self.name, self.props.get("Notification",""), NotificationType.INFO, action=[NotificationAction("Next Step", self.NextStep)])
            await self.push_update()

    async def on_timer_update(self,timer, seconds):
        await self.push_update()

    async def on_start(self):
        self.summary=""
        self.AutoNext = False if self.props.get("AutoNext", "No") == "No" else True
        if self.timer is None:
            self.timer = Timer(1 ,on_update=self.on_timer_update, on_done=self.on_timer_done)
        await self.push_update()

    async def on_stop(self):
        await self.timer.stop()
        self.summary = ""
        await self.push_update()

    async def run(self):
        while self.running == True:
            await asyncio.sleep(1)
            if self.timer.is_running is not True:
                self.timer.start()
                self.timer.is_running = True

        return StepResult.DONE

@parameters([Property.Number(label="Temp", configurable=True),
             Property.Sensor(label="Sensor"),
             Property.Kettle(label="Kettle"),
             Property.Text(label="Notification",configurable = True, description = "Text for notification when Temp is reached"),
             Property.Select(label="AutoMode",options=["Yes","No"], description="Switch Kettlelogic automatically on and off -> Yes")])
class MashInStep(CBPiStep):

    async def NextStep(self, **kwargs):
        await self.next()

    async def on_timer_done(self,timer):
        self.summary = ""
        self.kettle.target_temp = 0
        await self.push_update()
        if self.AutoMode == True:
            await self.setAutoMode(False)
        self.cbpi.notify(self.name, self.props.get("Notification","Target Temp reached. Please add malt and klick next to move on."), action=[NotificationAction("Next Step", self.NextStep)])

    async def on_timer_update(self,timer, seconds):
        await self.push_update()

    async def on_start(self):
        self.AutoMode = True if self.props.get("AutoMode","No") == "Yes" else False
        self.kettle=self.get_kettle(self.props.get("Kettle", None))
        if self.kettle is not None:
            self.kettle.target_temp = int(self.props.get("Temp", 0))
        if self.AutoMode == True:
            await self.setAutoMode(True)
        self.summary = "Waiting for Target Temp"
        if self.cbpi.kettle is not None and self.timer is None:
            self.timer = Timer(1 ,on_update=self.on_timer_update, on_done=self.on_timer_done)
        await self.push_update()

    async def on_stop(self):
        await self.timer.stop()
        self.summary = ""
        if self.AutoMode == True:
            await self.setAutoMode(False)
        await self.push_update()

    async def run(self):
        while self.running == True:
           await asyncio.sleep(1)
           sensor_value = self.get_sensor_value(self.props.get("Sensor", None)).get("value")
           if sensor_value >= int(self.props.get("Temp",0)) and self.timer.is_running is not True:
               self.timer.start()
               self.timer.is_running = True
        await self.push_update()
        return StepResult.DONE

    async def reset(self):
        self.timer = Timer(1 ,on_update=self.on_timer_update, on_done=self.on_timer_done)

    async def setAutoMode(self, auto_state):
        try:
            if (self.kettle.instance is None or self.kettle.instance.state == False) and (auto_state is True):
                await self.cbpi.kettle.toggle(self.kettle.id)
            elif (self.kettle.instance.state == True) and (auto_state is False):
                await self.cbpi.kettle.stop(self.kettle.id)
            await self.push_update()

        except Exception as e:
            logging.error("Failed to switch on KettleLogic {} {}".format(self.kettle.id, e))


@parameters([Property.Number(label="Timer", description="Time in Minutes", configurable=True), 
             Property.Number(label="Temp", configurable=True),
             Property.Sensor(label="Sensor"),
             Property.Kettle(label="Kettle"),
             Property.Select(label="AutoMode",options=["Yes","No"], description="Switch Kettlelogic automatically on and off -> Yes")])
class MashStep(CBPiStep):

    @action("Start Timer", [])
    async def start_timer(self):
        if self.timer.is_running is not True:
            self.cbpi.notify(self.name, 'Timer started', NotificationType.INFO)
            self.timer.start()
            self.timer.is_running = True
        else:
            self.cbpi.notify(self.name, 'Timer is already running', NotificationType.WARNING)

    @action("Add 5 Minutes to Timer", [])
    async def add_timer(self):
        if self.timer.is_running == True:
            self.cbpi.notify(self.name, '5 Minutes added', NotificationType.INFO)
            await self.timer.add(300)       
        else:
            self.cbpi.notify(self.name, 'Timer must be running to add time', NotificationType.WARNING)


    async def on_timer_done(self,timer):
        self.summary = ""
        self.kettle.target_temp = 0
        if self.AutoMode == True:
            await self.setAutoMode(False)
        self.cbpi.notify(self.name, 'Step finished', NotificationType.SUCCESS)
       
        await self.next()

    async def on_timer_update(self,timer, seconds):
        self.summary = Timer.format_time(seconds)
        await self.push_update()

    async def on_start(self):
        self.AutoMode = True if self.props.get("AutoMode", "No") == "Yes" else False
        self.kettle=self.get_kettle(self.props.Kettle)
        if self.kettle is not None:
            self.kettle.target_temp = int(self.props.get("Temp", 0))
        if self.AutoMode == True:
            await self.setAutoMode(True)
        await self.push_update()

        if self.cbpi.kettle is not None and self.timer is None:
            self.timer = Timer(int(self.props.get("Timer",0)) *60 ,on_update=self.on_timer_update, on_done=self.on_timer_done)
        elif self.cbpi.kettle is not None:
            try:
                if self.timer.is_running == True:
                    self.timer.start()
            except:
                pass

        self.summary = "Waiting for Target Temp"
        await self.push_update()

    async def on_stop(self):
        await self.timer.stop()
        self.summary = ""
        if self.AutoMode == True:
            await self.setAutoMode(False)
        await self.push_update()

    async def reset(self):
        self.timer = Timer(int(self.props.get("Timer",0)) *60 ,on_update=self.on_timer_update, on_done=self.on_timer_done)

    async def run(self):
        while self.running == True:
            await asyncio.sleep(1)
            sensor_value = self.get_sensor_value(self.props.get("Sensor", None)).get("value")
            if sensor_value >= int(self.props.get("Temp",0)) and self.timer.is_running is not True:
                self.timer.start()
                self.timer.is_running = True
                estimated_completion_time = datetime.fromtimestamp(time.time()+ (int(self.props.get("Timer",0)))*60)
                self.cbpi.notify(self.name, 'Timer started. Estimated completion: {}'.format(estimated_completion_time.strftime("%H:%M")), NotificationType.INFO)
        return StepResult.DONE

    async def setAutoMode(self, auto_state):
        try:
            if (self.kettle.instance is None or self.kettle.instance.state == False) and (auto_state is True):
                await self.cbpi.kettle.toggle(self.kettle.id)
            elif (self.kettle.instance.state == True) and (auto_state is False):
                await self.cbpi.kettle.stop(self.kettle.id)
            await self.push_update()

        except Exception as e:
            logging.error("Failed to switch on KettleLogic {} {}".format(self.kettle.id, e))


@parameters([Property.Number(label="Timer", description="Time in Minutes", configurable=True)])
class WaitStep(CBPiStep):

    async def on_timer_done(self, timer):
        self.summary = ""
        await self.next()

    async def on_timer_update(self, timer, seconds):
        self.summary = Timer.format_time(seconds)
        await self.push_update()

    async def on_start(self):
        if self.timer is None:
            self.timer = Timer(int(self.props.Timer) * 60, on_update=self.on_timer_update, on_done=self.on_timer_done)
        self.timer.start()

    async def on_stop(self):
        await self.timer.stop()
        self.summary = ""
        await self.push_update()

    async def reset(self):
        self.timer = Timer(int(self.props.Timer) * 60, on_update=self.on_timer_update, on_done=self.on_timer_done)

    async def run(self):
        while self.running == True:
            await asyncio.sleep(1)
        return StepResult.DONE


@parameters([Property.Select(label="toggle_type", options=["On", "Off"],description="Choose if Actor should be switched on or off in this step"),
             Property.Actor(label="Actor",description="Actor that should be toggled during this step")])
class ToggleStep(CBPiStep):
    async def on_timer_done(self, timer):
        self.summary = ""
        await self.next()

    async def on_timer_update(self, timer, seconds):
        self.summary = Timer.format_time(seconds)
        await self.push_update()

    async def on_start(self):
        if self.timer is None:
            self.timer = Timer(1, on_update=self.on_timer_update, on_done=self.on_timer_done)
        self.timer.start()
        self.type=self.props.get("toggle_type","Off")
        self.Actor=self.props.get("Actor", None)
        if self.Actor is not None and self.type == "On":
            await self.actor_on(self.Actor)
        if self.Actor is not None and self.type == "Off":
            await self.actor_off(self.Actor)


    async def on_stop(self):
        await self.timer.stop()
        self.summary = ""
        await self.push_update()

    async def reset(self):
        self.timer = Timer(1, on_update=self.on_timer_update, on_done=self.on_timer_done)

    async def run(self):
        while self.running == True:
            await asyncio.sleep(1)
        return StepResult.DONE

@parameters([Property.Number(label="Timer", description="Time in Minutes", configurable=True),
             Property.Actor(label="Actor")])
class ActorStep(CBPiStep):
    async def on_timer_done(self, timer):
        self.summary = ""
        await self.next()

    async def on_timer_update(self, timer, seconds):
        self.summary = Timer.format_time(seconds)
        await self.push_update()

    async def on_start(self):
        if self.timer is None:
            self.timer = Timer(int(self.props.Timer) * 60, on_update=self.on_timer_update, on_done=self.on_timer_done)
        self.timer.start()
        await self.actor_on(self.props.Actor)

    async def on_stop(self):
        await self.actor_off(self.props.Actor)
        await self.timer.stop()
        self.summary = ""
        await self.push_update()

    async def reset(self):
        self.timer = Timer(int(self.props.Timer) * 60, on_update=self.on_timer_update, on_done=self.on_timer_done)

    async def run(self):
        while self.running == True:
            await asyncio.sleep(1)
        return StepResult.DONE


@parameters([Property.Number(label="Timer", description="Time in Minutes", configurable=True), 
             Property.Number(label="Temp", description="Boil temperature", configurable=True),
             Property.Sensor(label="Sensor"),
             Property.Kettle(label="Kettle"),
             Property.Select(label="LidAlert",options=["Yes","No"], description="Trigger Alert to remove lid if temp is close to boil"),
             Property.Select(label="AutoMode",options=["Yes","No"], description="Switch Kettlelogic automatically on and off -> Yes"),
             Property.Select("First_Wort", options=["Yes","No"], description="First Wort Hop alert if set to Yes"),
             Property.Number("Hop_1", configurable = True, description="First Hop alert (minutes before finish)"),
             Property.Number("Hop_2", configurable=True, description="Second Hop alert (minutes before finish)"),
             Property.Number("Hop_3", configurable=True, description="Third Hop alert (minutes before finish)"),
             Property.Number("Hop_4", configurable=True, description="Fourth Hop alert (minutes before finish)"),
             Property.Number("Hop_5", configurable=True, description="Fifth Hop alert (minutes before finish)"),
             Property.Number("Hop_6", configurable=True, description="Sixth Hop alert (minutes before finish)")])
class BoilStep(CBPiStep):

    @action("Start Timer", [])
    async def start_timer(self):
        if self.timer.is_running is not True:
            self.cbpi.notify(self.name, 'Timer started', NotificationType.INFO)
            self.timer.start()
            self.timer.is_running = True
        else:
            self.cbpi.notify(self.name, 'Timer is already running', NotificationType.WARNING)

    @action("Add 5 Minutes to Timer", [])
    async def add_timer(self):
        if self.timer.is_running == True:
            self.cbpi.notify(self.name, '5 Minutes added', NotificationType.INFO)
            await self.timer.add(300)       
        else:
            self.cbpi.notify(self.name, 'Timer must be running to add time', NotificationType.WARNING)

    async def on_timer_done(self,timer):
        self.summary = ""
        self.kettle.target_temp = 0
        if self.AutoMode == True:
            await self.setAutoMode(False)
        self.cbpi.notify(self.name, 'Boiling completed', NotificationType.SUCCESS)
        await self.next()

    async def on_timer_update(self,timer, seconds):
        self.summary = Timer.format_time(seconds)
        self.remaining_seconds = seconds
        await self.push_update()

    async def on_start(self):

        self.lid_temp = 95 if self.get_config_value("TEMP_UNIT", "C") == "C" else 203
        self.lid_flag = True if self.props.get("LidAlert", "No") == "Yes" else False
        self.AutoMode = True if self.props.get("AutoMode", "No") == "Yes" else False
        self.first_wort_hop_flag = False 
        self.first_wort_hop=self.props.get("First_Wort", "No")
        self.hops_added=["","","","","",""]
        self.remaining_seconds = None

        self.kettle=self.get_kettle(self.props.get("Kettle", None))
        if self.kettle is not None:
            self.kettle.target_temp = int(self.props.get("Temp", 0))

        if self.cbpi.kettle is not None and self.timer is None:
            self.timer = Timer(int(self.props.get("Timer", 0)) *60 ,on_update=self.on_timer_update, on_done=self.on_timer_done)

        elif self.cbpi.kettle is not None:
            try:
                if self.timer.is_running == True:
                    self.timer.start()
            except:
                pass

        self.summary = "Waiting for Target Temp"
        if self.AutoMode == True:
            await self.setAutoMode(True)
        await self.push_update()

    async def check_hop_timer(self, number, value):
        if value is not None and self.hops_added[number-1] is not True:
            if self.remaining_seconds != None and self.remaining_seconds <= (int(value) * 60 + 1):
                self.hops_added[number-1]= True
                self.cbpi.notify('Hop Alert', "Please add Hop %s" % number, NotificationType.INFO)

    async def on_stop(self):
        await self.timer.stop()
        self.summary = ""
        self.kettle.target_temp = 0
        if self.AutoMode == True:
            await self.setAutoMode(False)
        await self.push_update()

    async def reset(self):
        self.timer = Timer(int(self.props.get("Timer", 0)) *60 ,on_update=self.on_timer_update, on_done=self.on_timer_done)

    async def run(self):
        if self.first_wort_hop_flag == False and self.first_wort_hop == "Yes":
            self.first_wort_hop_flag = True
            self.cbpi.notify('First Wort Hop Addition!', 'Please add hops for first wort', NotificationType.INFO)

        while self.running == True:
            await asyncio.sleep(1)
            sensor_value = self.get_sensor_value(self.props.get("Sensor", None)).get("value")
            
            if self.lid_flag == True and sensor_value >= self.lid_temp:
                self.cbpi.notify("Please remove lid!", "Reached temp close to boiling", NotificationType.INFO)
                self.lid_flag = False

            if sensor_value >= int(self.props.get("Temp", 0)) and self.timer.is_running is not True:
                self.timer.start()
                self.timer.is_running = True
                estimated_completion_time = datetime.fromtimestamp(time.time()+ (int(self.props.get("Timer", 0)))*60)
                self.cbpi.notify(self.name, 'Timer started. Estimated completion: {}'.format(estimated_completion_time.strftime("%H:%M")), NotificationType.INFO)
            else:
                for x in range(1, 6):
                    await self.check_hop_timer(x, self.props.get("Hop_%s" % x, None))

        return StepResult.DONE

    async def setAutoMode(self, auto_state):
        try:
            if (self.kettle.instance is None or self.kettle.instance.state == False) and (auto_state is True):
                await self.cbpi.kettle.toggle(self.kettle.id)
            elif (self.kettle.instance.state == True) and (auto_state is False):
                await self.cbpi.kettle.stop(self.kettle.id)
            await self.push_update()

        except Exception as e:
            logging.error("Failed to switch on KettleLogic {} {}".format(self.kettle.id, e))


@parameters([Property.Number(label="Temp", configurable=True, description="Target temperature for cooldown. Notification will be send when temp is reached and Actor can be triggered"),
             Property.Sensor(label="Sensor", description="Sensor that is used during cooldown"),
             Property.Actor(label="Actor", description="Actor can trigger a valve for the cooldwon to target temperature"),
             Property.Kettle(label="Kettle")])
class CooldownStep(CBPiStep):

    async def on_timer_done(self,timer):
        self.summary = ""
        if self.actor is not None:
            await self.actor_off(self.actor)
        self.cbpi.notify('CoolDown', 'Wort cooled down. Please transfer to Fermenter.', NotificationType.INFO)
        await self.next()

    async def on_timer_update(self,timer, seconds):
        await self.push_update()

    async def on_start(self):
        warnings.simplefilter('ignore', np.RankWarning)
        self.temp_array = []
        self.time_array = []
        self.kettle = self.get_kettle(self.props.get("Kettle", None))
        self.actor = self.props.get("Actor", None)
        self.target_temp = int(self.props.get("Temp",0))
        self.Interval = 15 # Interval in minutes on how often cooldwon end time is calculated

        self.cbpi.notify(self.name, 'Cool down to {}°'.format(self.target_temp), NotificationType.INFO)
        if self.timer is None:
            self.timer = Timer(1,on_update=self.on_timer_update, on_done=self.on_timer_done)
        self.start_time=time.time()
        self.temp_array.append(self.get_sensor_value(self.props.get("Sensor", None)).get("value"))
        self.time_array.append(time.time())
        self.next_check = self.start_time + self.Interval * 60
        self.count = 0
        self.initial_date = None

    async def on_stop(self):
        await self.timer.stop()
        self.summary = ""
        if self.actor is not None:
            await self.actor_off(self.actor)
        await self.push_update()

    async def reset(self):
        self.timer = Timer(1,on_update=self.on_timer_update, on_done=self.on_timer_done)

    async def run(self):
        timestring = datetime.fromtimestamp(self.start_time)
        if self.actor is not None:
            await self.actor_on(self.actor)
        self.summary="Started: {}".format(timestring.strftime("%H:%M"))
        await self.push_update()
        while self.running == True:
            current_temp = self.get_sensor_value(self.props.get("Sensor", None)).get("value")
            if self.count == 10:
                self.temp_array.append(current_temp)
                current_time = time.time()
                if self.initial_date == None:
                    self.initial_date = current_time 
                self.time_array.append(current_time)
                self.count = 0
            if time.time() >= self.next_check:
                self.next_check = time.time() + (self.Interval * 60)

                cooldown_model = np.poly1d(np.polyfit(self.temp_array, self.time_array, 2))
                target_time=cooldown_model(self.target_temp)
                target_timestring= datetime.fromtimestamp(target_time)
                self.summary="ECT: {}".format(target_timestring.strftime("%H:%M"))
                self.cbpi.notify("Cooldown Step","Current: {}°, reaching {}° at {}".format(round(current_temp,1), self.target_temp, target_timestring.strftime("%d.%m %H:%M")), NotificationType.INFO)
                await self.push_update()

            if current_temp <= self.target_temp and self.timer.is_running is not True:
                self.timer.start()
                self.timer.is_running = True

            self.count +=1
            await asyncio.sleep(1)

        return StepResult.DONE

def setup(cbpi):
    '''
    This method is called by the server during startup 
    Here you need to register your plugins at the server

    :param cbpi: the cbpi core 
    :return: 
    '''

    cbpi.plugin.register("MashInStep", MashInStep)
    cbpi.plugin.register("MashStep", MashStep)
    cbpi.plugin.register("BoilStep", BoilStep)
    cbpi.plugin.register("CooldownStep", CooldownStep)
    cbpi.plugin.register("WaitStep", WaitStep)
    cbpi.plugin.register("ToggleStep", ToggleStep)
    cbpi.plugin.register("ActorStep", ActorStep)
    cbpi.plugin.register("NotificationStep", NotificationStep)



from cbpi.controller.upload_controller import UploadController
from cbpi.api.dataclasses import Props, Step
from aiohttp import web
from cbpi.api import *
import logging

class UploadHttpEndpoints():

    def __init__(self, cbpi):
        self.cbpi = cbpi
        self.controller : UploadController = cbpi.upload
        self.cbpi.register(self, "/upload")


    @request_mapping(path='/', method="POST", auth_required=False)
    async def FileUpload(self, request):
        data = await request.post()
        await self.controller.FileUpload(data)
        return web.Response(status=200)




    @request_mapping(path='/kbh', method="GET", auth_required=False)
    async def get_kbh_list(self, request):
        """

        ---
        description: Get Recipe list from Kleiner Brauhelfer 
        tags:
        - Upload
        responses:
            "200":
                description: successful operation
        """

        kbh_list = await self.controller.get_kbh_recipes()
        return web.json_response(kbh_list)

    @request_mapping(path='/kbh', method="POST", auth_required=False)
    async def create_kbh_recipe(self, request):
        """

        ---
        description: Create Recipe from KBH database with selected ID
        tags:
        - Upload
        parameters:
        - name: "id"
          in: "body"
          description: "Recipe ID: {'id': ID}"
          required: true
          type: "string"
        responses:
            "200":
                description: successful operation
        """

        kbh_id = await request.json()
        await self.controller.kbh_recipe_creation(kbh_id['id'])
        return web.Response(status=200)

    @request_mapping(path='/xml', method="GET", auth_required=False)
    async def get_xml_list(self, request):
        """

        ---
        description: Get recipe list from xml file
        tags:
        - Upload
        responses:
            "200":
                description: successful operation
        """

        xml_list = await self.controller.get_xml_recipes()

        return web.json_response(xml_list)

    @request_mapping(path='/xml', method="POST", auth_required=False)
    async def create_xml_recipe(self, request):
        """

        ---
        description: Create recipe from xml file with selected id
        tags:
        - Upload
        parameters:
        - name: "id"
          in: "body"
          description: "Recipe ID: {'id': ID}"
          required: true
          type: "string"

        responses:
            "200":
                description: successful operation
        """

        xml_id = await request.json()
        await self.controller.xml_recipe_creation(xml_id['id'])
        return web.Response(status=200)

    @request_mapping(path='/json', method="GET", auth_required=False)
    async def get_json_list(self, request):
        """

        ---
        description: Get recipe list from json file
        tags:
        - Upload
        responses:
            "200":
                description: successful operation
        """

        json_list = await self.controller.get_json_recipes()

        return web.json_response(json_list)

    @request_mapping(path='/json', method="POST", auth_required=False)
    async def create_json_recipe(self, request):
        """

        ---
        description: Create recipe from json file with selected id
        tags:
        - Upload
        parameters:
        - name: "id"
          in: "body"
          description: "Recipe ID: {'id': ID}"
          required: true
          type: "string"

        responses:
            "200":
                description: successful operation
        """

        json_id = await request.json()
        await self.controller.json_recipe_creation(json_id['id'])
        return web.Response(status=200)

    @request_mapping(path='/bf/{offset}/', method="POST", auth_required=False)
    async def get_bf_list(self, request):
        """

        ---
        description: Get recipe list from Brewfather App
        tags:
        - Upload
        responses:
            "200":
                description: successful operation
        """
        offset = request.match_info['offset']
        bf_list = await self.controller.get_brewfather_recipes(offset)

        return web.json_response(bf_list)

    @request_mapping(path='/bf', method="POST", auth_required=False)
    async def create_bf_recipe(self, request):
        """

        ---
        description: Create recipe from Brewfather Web App with selected id
        tags:
        - Upload
        parameters:
        - name: "id"
          in: "body"
          description: "Recipe ID: {'id': ID}"
          required: true
          type: "string"

        responses:
            "200":
                description: successful operation
        """

        bf_id = await request.json()
        await self.controller.bf_recipe_creation(bf_id['id'])
        return web.Response(status=200)

    @request_mapping(path="/getpath", auth_required=False)
    async def http_getpath(self, request):
        
        """

        ---
        description: get path for recipe creation
        tags:
        - Upload
        responses:
            "200":
                description: successful operation
        """
        return  web.json_response(data=self.controller.get_creation_path())




from enum import Enum

class ConfigType(Enum):
    STRING = "string"
    NUMBER = "number"
    SELECT = "select"
    KETTLE = "kettle"
    ACTOR = "actor"
    SENSOR = "sensor"
    STEP = "step"
    FERMENTER = "fermenter"


